/* x.c - General parts of LibXPCE, the PseudoColor emulation library for X.
 *
 * Copyright (C) 2000 Boris Gjenero
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1, as published by the Free Software Foundation
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You may contact Boris Gjenero at bgjenero@sympatico.ca
 */

/* I originally started with i_video.c from the original unixdoom release
 * but practically everything has been totally rewritten, except:
 * - Some variable names
 * - grabsharedmemory
 * - x_SetNullCursor
 *
 * Some other programs are mentioned here because I found a neat idea there.
 * For example the root widow background setting is based on stuff in
 * xsetroot.
 */


/***************************** INCLUDES **************************************/

#include "xconfig.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h> 
#include <limits.h>  /* for LONG_MIN and LONG_MAX */
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
/* Had to dig up XShm.c for this one. */
/* It is in the libXext, but not in the XFree86 headers. */
#ifdef __linux__
int XShmGetEventBase( Display* dpy ); /* problems with g++? */
#endif
#endif

#ifdef XF86DGA
#include <X11/extensions/xf86dga.h>
#endif

#include "x.h"
#include "libxpce.h"

/* Define TIME_XLAT to some number to run the translator that number of
 * times and report the execution time.
 */
#define TIME_XLAT 10
#undef TIME_XLAT

/************************* GLOBAL VARIABLES **********************************/

/*** Various mostly self explanatory X things ***/
Display*	X_display = NULL;
Window		X_mainWindow = None;
Colormap	X_cmap = None;
Visual*		X_visual;
XVisualInfo*    X_visualinfo = NULL;
GC		X_gc = None;
int		X_screen;
Screen         *X_scr;
XImage*		X_image = NULL;
char*           X_displayname;
XColor	        X_colors[INPAL_SIZE];
Window          X_rootwin;
Colormap        X_oldrcmap;   /* Old root colormap */
int             X_error;      /* Flag set by error flagging error handler */
int             X_screenwidth;
int             X_screenheight;
char		*X_caption = "LibXPCE";

/*** Atoms ***/

/* Atoms for WM_PROTOCOLS */
char *X_wmpatomnames[] = {
  "WM_PROTOCOLS",
  "WM_DELETE_WINDOW"
};
Atom X_wmpatoms[] = { 0, 0 };
#define X_atom_WM_PROTOCOLS (X_wmpatoms[0])
#define X_atom_WM_DELETE_WINDOW (X_wmpatoms[1])

/* Atoms for root window background setting */
Atom Xa__XSETROOT_ID = 0;
char *Xan__XSETROOT_ID = "_XSETROOT_ID";

#ifdef MITSHM
/*** MIT-SHM-only variables ***/
XShmSegmentInfo X_shminfo;
int             X_shmeventtype;
int             doShm;
int             shmFinished;
int             X_disableshm = 0;
#endif

/*** Position variables ***/

/* Position has been remembered */
int             X_psrem = 0;

/* We are getting sythetic ConfigureNotify events.  Ignore position in
   non-synthetic ConfigureNotify events. */
int             X_cnpoison = 0;

/* Saved position from last time when we were in a normal window */
int             X_savedx = 0;
int             X_savedy = 0;

/* Saved width and height from last time we were in a normal window */
int             X_savedwidth = DEFWIDTH;
int             X_savedheight = DEFHEIGHT;

/* Current window width and height */
int		X_width;
int		X_height;

/* Mouse position */
/* - Use MotionNotify instead of XQueryPointer */
int             X_trackmouse = 0;
/* - Have a position through MotionNotify */
int             X_havemouse = 0;
/* - Actual Position */
int             X_mousex;
int             X_mousey;

/*** Misc. variables ***/

/* Pointer to the graphics buffer used by acidwarp.  This may be 
 * X_image->data or it may be separate memory if some translation is 
 * necessary.
 */
unsigned char *buf_graf;

/* Set to 1 if we have a changeable colormap to use and worry about.
 * Otherwise 0.  This doesn't indicate whether we're actually changing
 * it actively or whether the image is being redrawn at each colormap
 * change.
 */
int X_colormapped;

/* If 1 when we need to change the image we just use XClearArea on the whole
 * window.  The image then actually gets drawn in response to Expose events.
 * This way we don't repaint obscured areas.
 * If 0 the image is immediately updated to the X server.  Use this if 
 * XClearArea insists on wasting time and causing flicker by clearing
 * the window.
 */
int X_useclear; 

/* Set to 1 after the 1st expose event with 0 count */
int X_oktodraw = 0;

/* This is the number of palette entries that we have allocated.  It may
 * be controlled by the -c option.
 */
int X_u_numcolors = INPAL_SIZE;   /* User specified */
int X_numcolors = 0;              /* Current */

/* This gets set to 1 if a window manager is detected
 */
int X_wmdetected = 0;

/* If this is non-zero, XAllocColor will be used instead of our much
   faster routines (see x.h) */
int X_xalloccolor = 0;

/* If this is non-zero, XPutPixel will be used to draw into the XImage */
int X_xputpixel = 0;

/* Display mode specifications (see x.h) */
int X_dmode = DMODE_UNSPECIFIED;

/* Full screen display mode (see x.h) */
int X_fsmode = FSMODE_DGA;

/* Basic X initialization done */
int X_xxed = 0;

/* Error handlers installed */
int X_ehinst = 0;

/* The window and other associated stuff has been set */
int X_windowedx = 0;

/* Need to retain resources (eg. if setting the background) */
int X_retainres = 0;

/* Track window position */
int X_trackpos;

#ifdef NO_POINTLESS_STORECOLORS
/* True if our colormap is installed */
int X_cmapin = 0;

/* True if our colormap is up to date */
int X_cmapupd = 1;
#endif

#ifdef OBSCURED_SLEEP
/* Visibility state of main window from last VisibilityNotify event */
int X_visstate = VisibilityFullyObscured;

/* Main window currently mapped */
int X_mwmapped = 0;
#endif

/* If this is set we should get back to Acidwarp so it can quit */
volatile int X_quitflag;

/* Retain last image as root window background */
int X_setbg = 0;

/* Server currently grabbed */
int X_servergrabbed = 0;

#ifdef COLORMAP_THEFT
/* We want to grab the colormap.  (see x.h) */
int X_grabcmap = GRABCMAP_NO;
/* Grab the colormap only in full screen or root mode */
int X_grabcmapfsr = 0;
/* Colormap currently grabbed with WM_COLORMAP_NOTIFY */
int X_colormapgrabbed = 0;
#if defined(MAX_CMAP_REINSTALLS) && MAX_CMAP_REINSTALLS > 0
int X_cmapreinstallcnt = 0;
#endif
#endif

/* What the user wants for the colormap (see x.h) */
int X_cmappref = CMAPPREF_UNSPEC; 

/* If 1 we use a private colormap in colormapped modes.  If 0 we use the
 * default colormap.
 */
int X_privatecmap = PRIVATECMAP;

/* Bits per pixel in the final data destination */
unsigned int X_bits_per_pixel;

/* Bytes per line in the final destination */
unsigned int X_bytes_per_line;

/* Scanline pad in the XImage */
unsigned int X_scanline_pad;

/* Byte order in the final destination (MSBFirst or LSBFirst) */
unsigned int X_byte_order;

/************************** GENERAL FUNCTIONS *******************************/

/* x_Error - Report error and exit */
void x_Error(const char *s) {
  fprintf(stderr, s);
  exit(-1);
}

/* If server isn't grabbed, grab it.  Increment grab count */
void x_GrabServer() {
  if (!X_servergrabbed) {
    XGrabServer(X_display);
  }
  X_servergrabbed++;
}

/* Decrement server grab count.  If it is 0, ungrab server */
void x_UngrabServer() {
  if (X_servergrabbed > 0) {
    X_servergrabbed--;
    if (!X_servergrabbed) {
      XUngrabServer(X_display);
    }
  }
}

/* x_GetPixmapFormat - This sets X_bits_per_pixel and X_scanline_pad from
 *                     a pixmap format of the desired depth.  This was
 *                     taken from SDL 1.6.
 */
int x_GetPixmapFormat(int depth) {
  XPixmapFormatValues *pix_format;
  int i, num_formats;

  /* Get info on pixmap formats */
  pix_format = XListPixmapFormats(X_display, &num_formats);
  if ( pix_format == NULL ) {
    x_Error("Couldn't determine Pixmap formats");
  }

  for ( i=0; i<num_formats; ++i ) {
    if ( depth == pix_format[i].depth )
      break;
  }

  if ( i != num_formats ) {
    X_bits_per_pixel = pix_format[i].bits_per_pixel;
    X_scanline_pad =  pix_format[i].scanline_pad;
    return 0;
  } else {
    fprintf(stderr, "warning:  Failed to find pixmap format for the depth "
	    "that will be used.\n");	   
    X_bits_per_pixel = depth;
    X_scanline_pad = 8;
    return -1;
  }

  XFree((char *)pix_format);
}


#ifdef MITSHM

/* grabsharedmemory function was taken from unixdoom's i_video.c.  It may
 * be total overkill for grabbing shared memory but it is useful 
 * during development if the program crashes.
 */

/* grabsharedmemory - Grab a shared memory area */
void *grabsharedmemory(int size) {
  int                   key = MITSHM_DEFKEY;
  struct shmid_ds       shminfo;
  int                   minsize = 320*200;
  int                   id;
  int                   rc;
  /* UNUSED int done=0; */
  int                   pollution=5;
  void                 *ptr;
  
  /* try to use what was here before */
  do
  {
    id = shmget((key_t) key, minsize, 0777); /* just get the id */
    if (id != -1)
    {
      rc=shmctl(id, IPC_STAT, &shminfo); /* get stats on it */
      if (!rc) 
      {
        if (shminfo.shm_nattch)
        {
	  if (kill(shminfo.shm_cpid, 0)) {
	    fprintf(stderr, "warning:  Found shared memory from a non-"
		    "existant copy of xacidwarp.\n"
		    "          Some process is still attached to it."
		    "  That process is probably\n"
		    "          your X server.  You may need to restart X to "
		    "free that shared memory.\n"
		    "          (id = %i, nonexistant PID = %i)\n", 
		    id, shminfo.shm_cpid);
	  }
	  key++;
        }
        else
        {
          if (getuid() == shminfo.shm_perm.cuid)
          {
            rc = shmctl(id, IPC_RMID, 0);
            if (!rc)
              fprintf(stderr,
                      "Was able to kill my old shared memory\n");
            else
              x_Error("Was NOT able to kill my old shared memory\n");
            
            id = shmget((key_t)key, size, IPC_CREAT|0777);
            if (id==-1)
              x_Error("Could not get shared memory\n");
            
            rc=shmctl(id, IPC_STAT, &shminfo);
            
            break;
            
          }
          if (size >= shminfo.shm_segsz)
          {
            fprintf(stderr,
                    "will use %d's stale shared memory\n",
                    shminfo.shm_cpid);
            break;
          }
          else
          {
            fprintf(stderr,
                    "warning: can't use stale "
                    "shared memory belonging to id %d, "
                    "key=0x%x\n",
                    shminfo.shm_cpid, key);
            key++;
          }
        }
      }
      else
      {
        fprintf(stderr, "could not get stats on key=%d\n", key);
      }
    }
    else
    {
      id = shmget((key_t)key, size, IPC_CREAT|0777);
      if (id==-1)
      {
        perror("shmget");
        x_Error("Could not get any shared memory\n");
      }
      break;
    }
  } while (--pollution);
  
  if (!pollution)
  {
    x_Error("Sorry, system too polluted with stale "
            "shared memory segments.\n");
    }   
  
  X_shminfo.shmid = id;
  
  /* attach to the shared memory segment */
  ptr = shmat(id, 0, 0);
  
  if (!ptr)
    {
      perror("shmat");
      x_Error("shmat() failed in grabsharedmemory()\n");
    }
  
  /* fprintf(stderr, "shared memory id=%d, addr=0x%x\n", id,
     (int) (ptr)); */
#ifdef MITSHM_KEYWARN
  if (key != MITSHM_DEFKEY) {
    fprintf(stderr, "warning:  Using non-default shared memory key.  In case "
	    "of a crash, shared\n          memory won't be recovered "
	    "automatically.\n");
  }
#endif
  return ptr;
}
#endif

/* x_CreateImage() - Creates the image, either normally or as shared memory */
void x_CreateImage() {
  XImage *image;

#ifdef MITSHM
  if (doShm) {
    /* Create an image in shared memory */

    X_shmeventtype = XShmGetEventBase(X_display) + ShmCompletion;
    
    /* Create the image */
    image = XShmCreateImage(X_display,
			      X_visual,
			      X_visualinfo->depth,
			      ZPixmap,
			      0,
			      &X_shminfo,
			      X_width,
			      X_height );
    
    if (image == NULL)
      x_Error("Failed to create X SHM image\n");
    
    /* Grab the shared memory */
    image->data = X_shminfo.shmaddr = 
      grabsharedmemory(image->bytes_per_line * image->height);
    
    /* Get the X server to attach to it */
    if (!XShmAttach(X_display, &X_shminfo))
      x_Error("XShmAttach() failed in InitGraphics()\n");

  } else {
#endif
    /* We need this to get scanline pad.  Perhaps if that was wrong we
       wouldn't be able to create the image. */
    x_GetPixmapFormat(X_visualinfo->depth);

    /* Create an ordinary XImage */
    image = XCreateImage(X_display,
			 X_visual,
			 X_visualinfo->depth,
			 ZPixmap,
			 0,
			 NULL,
			 X_width, X_height,
			 X_scanline_pad,
			 0);

    if (image == NULL)
      x_Error("Failed to create X image\n");

    /* Allocate data for image */
    image->data = calloc(image->bytes_per_line * image->height, 1);
    if (!image->data)
      x_Error("Failed to allocate memory for X image.\n");
#ifdef MITSHM
  }
#endif

  /* Allocate separate acidwarp graphics buffer if necessary */
  X_needxlat &= ~NEEDXLAT_BPL;
  if (X_needxlat == 0) {
    if (image->bits_per_pixel != 8) {
      X_needxlat |= NEEDXLAT_CMAP;
    } else if (image->bytes_per_line != X_width) {
      X_needxlat |= NEEDXLAT_BPL;
    } else {
      /* Okay, under those conditions we don't need mapping */
      buf_graf = image->data;
    }
  }

#ifndef NO_BUFGRAF
  if (X_needxlat) {
    /* If we need translation allocate a separate Acidwarp image buffer */
    buf_graf = calloc(X_width * X_height, 1);
    if (!buf_graf)
      x_Error("Failed to allocate memory acidwarp image.\n");
  }
#endif

  X_red_mask = image->red_mask;
  X_green_mask = image->green_mask;
  X_blue_mask = image->blue_mask;
  X_bits_per_pixel = image->bits_per_pixel;
  X_bytes_per_line = image->bytes_per_line;
  X_byte_order = image->byte_order;

  /* At this point the image is okay */
  X_image = image;

  X_xlatdest = image->data;
  if (X_needxlat) {
    x_SelectXlatFunc();
  }
}

/* x_DestroyImage() - Destroy the X image we have allocated. */
void x_DestroyImage() {
  XImage *image;
  
  image = X_image;
  X_image = NULL;

#ifndef NO_BUFGRAF
  /* If there was a separate graphics buffer, free that. */
  if ((void *)buf_graf != (void *)image->data)
    free (buf_graf);
#endif

#ifdef MITSHM
  if (doShm) {
    /* If we are using shared memory we have to detach and deallocate it */
    
    /* Detach from X server */
    if (X_display != NULL) {
      if (!XShmDetach(X_display, &X_shminfo))
	x_Error("XShmDetach() failed in x_DestroyImage()\n");
      /* An XSync here insures that the server detaches *now*.  Otherwise
	 recreating the shared memory segment after a resize may fail */
      XSync(X_display, False);
    }
    
    /* Release shared memory. */
    shmdt(X_shminfo.shmaddr);
    shmctl(X_shminfo.shmid, IPC_RMID, 0);

    /* Since there is no data to free in the image this is the way to 
       get rid of it */
    XFree(image);
  } else {
#endif
    /* If we aren't using shared memory then just destroy the image.
       This will free the data too. */
    XDestroyImage(image);
#ifdef MITSHM
  }
#endif
}

/* x_PutImagePart - Puts a part of the image to the screen */
void x_PutImagePart(Drawable d, int x, int y, int w, int h) {
#ifdef MITSHM
  if (doShm) {
    if (!XShmPutImage(X_display,
		      d,
		      X_gc,
		      X_image,
		      x, y,
		      x, y,
		      w, h,
#ifdef DOSYNC
		      True 
#else
		      False
#endif
		      ))
      x_Error("XShmPutImage() failed\n");
	  
#ifdef DOSYNC
    /* wait for it to finish and processes all input events */
    shmFinished = 0;
    do
      {
	x_GetEvent();
      } while (!shmFinished);
#endif 
  } else {
#endif
    XPutImage(X_display,
	      d,
	      X_gc,
	      X_image,
	      x, y,
	      x, y,
	      w, h);
#ifdef DOSYNC	    
    /* sync up with server */
    XSync(X_display, False);
#endif DOSYNC
#ifdef MITSHM
  }
#endif
}

void x_InitWindow(void);
void x_CloseWindow(void);
void x_GrabColormap(int);
void x_GrabRootCMap(int);

void xpce_ToggleFullscreen() {
  if (geteuid()) {
	  fprintf(stderr, "Must be root for XF86DGA, please run as root for fullscreen\n");
	  return;
  }

  /* Save previous window height so it can be restored when we are
   * windowed again 
   */
  if (!(X_dmode & DMODE_FULLSCREEN)) {
    X_savedwidth = X_width;
    X_savedheight = X_height;
    X_psrem = 1;
  }

  if (
#ifdef XF86DGA
      X_fsmode != FSMODE_DGA && 
#endif
      !((X_visdeftype & VISDEF_FSHAND) && (X_visdeftype & VISDEF_HANDLER) &&
	X_visdef_handler != X_visdef_fshand)
#ifdef COLORMAP_THEFT
      && !(X_fsmode == FSMODE_OR
	&& X_cmappref == CMAPPREF_UNSPEC && X_colormapped &&
	!X_grabcmap)
#endif
      ) {
    /* If the full-screen mode isn't DGA mode and we haven't defined 
     * different handlers for windowed and full screen modes then
     * we don't have to restart the graphics driver
     */
    /* Destroy and recreate our main window */
    x_DestroyImage();
    x_CloseWindow();
    
    X_dmode ^= DMODE_FULLSCREEN;
    
#ifdef COLORMAP_THEFT
    if (X_grabcmap != 0 && X_grabcmapfsr != 0) {
      /* Change colormap grab state */
      if (X_dmode & DMODE_FULLSCREEN) {
	if (X_grabcmap & GRABCMAP_ICCCM)
	  x_GrabColormap(1);
	if (X_grabcmap & GRABCMAP_ROOT)
	  x_GrabRootCMap(1);
      } else {
	if (X_grabcmap & GRABCMAP_ICCCM)
	  x_GrabColormap(0);
	if (X_grabcmap & GRABCMAP_ROOT)
	  x_GrabRootCMap(0);
      }
    }
#endif

    x_InitWindow();
    x_CreateImage();
    xpceC_HandleResize(X_width, X_height);

    /* Note:  If we were using full screen window mode the window could
     *        have been reused.  But I was getting expose events for the
     *        old size even after the ConfigureNotify event and so
     *        I'm always killing the window.
     */
  } else {
    /* If the full-screen mode uses DGA or if we will be switching between
     * a private and a public colormap then we really do need to restart
     * the graphics driver 
     */
    fprintf(stderr, "Restarting graphics driver due to full-screen toggle.\n");
    xpce_CloseGraph();
    X_dmode ^= DMODE_FULLSCREEN;
    xpce_InitGraph();
    xpceC_HandleResize(X_width, X_height);
  }
}

/* x_DispatchKeyEvent - Translates X key presses to Acidwarp commands */
void x_DispatchKeyEvent(XKeyEvent *xevent) {
  int i, n;
  char buf[5];
  KeySym k;

  if (!X_quitflag) {
    n = XLookupString(xevent, buf, 4, &k, NULL);
    switch(k) {
    case XK_Return:
    case XK_KP_Enter:
      if ((xevent->state & (Mod1Mask | ControlMask)) 
	  && !(X_dmode & DMODE_ROOT)) {
	/* I wanted Alt-Enter for the full screen mode toggle key, but
         * unfortunately Enlightenment treats it specially.  So I'm also
         * supporting Ctrl-Enter.  I hope that at least one of those will
	 * work with all window managers.
	 */
	xpce_ToggleFullscreen();
      }
      break;
    default:
      if (n > 0) {
	for (i = 0; i < n; i++) {
	  xpceC_HandleInputChar(buf[i]);
	}
      } else {
	xpceC_HandleInputKeySym(k);
      }
      break;
    }
  }
}

/* x_GetEvent - gets next X event and processes it */
void x_GetEvent(void)
{
  XEvent xevent;

  XNextEvent(X_display, &xevent);
  
  /* DANGER:  Events stay queued.  It's possible for an event to
   *          come at a totally unexpected time, like for example
   *          an Expose while running in DGA mode, or a KeyRelease
   *          in normal mode.
   */
  
  switch (xevent.type) {
  case KeyPress:
    /* KeyPress - Send keyboard command to Acidwarp */
    x_DispatchKeyEvent(&xevent.xkey);
    break;
    
  case Expose:
    /* Note:  We most definitely don't want to handle expose events
     *        left behind from an old window.
     */
    if (
#ifdef XF86DGA
	!X_dga &&
#endif
	xevent.xexpose.window == X_mainWindow) {
      if (xevent.xexpose.x < 0 || xevent.xexpose.y < 0 || 
	  (xevent.xexpose.width + xevent.xexpose.x) > X_width ||
	  (xevent.xexpose.height + xevent.xexpose.y) > X_height) {
	fprintf(stderr, "Bad expose event: syn=%u lser=%lu win=%lu"
		"%i,%i %ix%i\n", 
		xevent.xexpose.send_event, 
		xevent.xexpose.serial,
		xevent.xexpose.window,
		xevent.xexpose.x, xevent.xexpose.y, 
		xevent.xexpose.width, xevent.xexpose.height);
      } else {
	/* Expose - Send that part of the image to the X server */
	x_PutImagePart(X_mainWindow, xevent.xexpose.x, xevent.xexpose.y,
		       xevent.xexpose.width, xevent.xexpose.height);
	
	/* Set okay to draw flag */
	if (xevent.type == Expose && !xevent.xexpose.count) {
	  X_oktodraw = 1;
	}
      }
    }
    break;
    
  case ConfigureNotify:
#ifdef XF86DGA
    if (!X_dga) {
#endif
      /* If a window manager isn't running you only get real events.  If
       * one is running you typically become reparented and you get
       * the real ConfigureNotify (with reparented x,y) sometimes followed
       * by a synthetic ConfigureNotify with the real x and y. 
       * twm sends a synthetic ConfigureNotify for moving, but not for 
       * resizing.
       * Also note that you don't get the initial ConfigureNotify when
       * the window is first mapped, but with a window manager you
       * will get one when it reparents you.  
       */
#if 0
      fprintf(stderr, "ConfigureNotify %i,%i %ix%i\n", 
	      xevent.xconfigure.x, 
	      xevent.xconfigure.y,
	      xevent.xconfigure.width,
	      xevent.xconfigure.height);
#endif
      
      if (xevent.xconfigure.send_event) {
	X_cnpoison = 1;
      }
      
      if (!(X_dmode & DMODE_FULLSCREEN) &&
	  (xevent.xconfigure.width == X_screenwidth &&
	   xevent.xconfigure.height == X_screenheight &&
	   xevent.xconfigure.x == 0 &&
	   xevent.xconfigure.y == 0)) {
	/* Entering full screen mode. */
	X_trackpos = 0;
	X_dmode |= DMODE_FULLSCREEN;
	fprintf(stderr, "FULLSCREEN\n");
      }
      
      if (X_width != xevent.xconfigure.width ||
	  X_height != xevent.xconfigure.height) {
	/* It's only a resize if the window size actually changes */
	
	/* Destroy the old image */
	x_DestroyImage();
	
	/* Update width and height.  (Note that in fact either or both
	   may have changed) */
	X_width = xevent.xconfigure.width;
	X_height = xevent.xconfigure.height;
	
	/* Create the new one */
	x_CreateImage();
	
	/* Now acidwarp has to handle the resize */
	xpceC_HandleResize(X_width, X_height);
	
	if ((X_dmode & DMODE_FULLSCREEN) && 
	    (X_width != X_screenwidth || X_height != X_screenheight)) {
	  /* Hmm... we're not full screen anymore. */
	  /* The problem with position is that it may be different, so
	   * we're just relying on the size.
	   */
	  X_dmode &= ~DMODE_FULLSCREEN;
	  X_trackpos = 1;
	}
      }
      
      /* Save our position for full screen toggling */
      if (X_trackpos &&
	  !(X_cnpoison && !xevent.xconfigure.send_event)) {
	X_savedx = xevent.xconfigure.x;
	X_savedy = xevent.xconfigure.y;
      }
#ifdef XF86DGA
    }
#endif
    break;
    
#if (defined(COLORMAP_THEFT) && (!defined(MAX_CMAP_REINSTALLS) \
|| MAX_CMAP_REINSTALLS > 0)) || defined(NO_POINTLESS_STORECOLORS)
  case ColormapNotify:
#ifdef XF86DGA
    if (!X_dga) {
#endif
      /* Change of colormap */
      if (xevent.xcolormap.new == False && 
	  xevent.xcolormap.colormap == X_cmap &&
	  xevent.xcolormap.window == X_mainWindow) {
	/* Okay, this event does concern us */
	if (xevent.xcolormap.state == ColormapUninstalled) {
	    /* Our colormap has been uninstalled */
#ifdef NO_POINTLESS_STORECOLORS	  
	    X_cmapin = 0;
#endif
#if defined(COLORMAP_THEFT) && (!defined(MAX_CMAP_REINSTALLS) \
|| MAX_CMAP_REINSTALLS > 0)
	    if (X_colormapgrabbed & GRABCMAP_ICCCM) {
	      /* Colormap should be grabbed */
	      /* Hmm... someone is trying to uninstall our colormap */
#ifdef MAX_CMAP_REINSTALLS
	      if (X_cmapreinstallcnt < MAX_CMAP_REINSTALLS) {
#endif
		XInstallColormap(X_display, X_cmap);
#ifdef MAX_CMAP_REINSTALLS
		X_cmapreinstallcnt++;
		if (X_cmapreinstallcnt == MAX_CMAP_REINSTALLS) {
		  fprintf(stderr, 
		       "The maximum number of colormap reinstalls has been"
		       " reached.\nAcidwarp is probably fighting another "
		       "X client, (probably the window\nmanager) for control "
		       "of the colormap.  Further colormap reinstalls have\n"
		       "been disabled.  (See the MAX_CMAP_REINSTALLS compile-"
		       "time option.)\n");
		}
	      }
#endif
	    }
#endif
	}
#ifdef NO_POINTLESS_STORECOLORS
	else {
	  /* Colormap has been installed */
	  X_cmapin = 1;
	  
	  /* If the colors haven't been updated then start update them */
	  if (!X_cmapupd) {
	    XStoreColors(X_display, X_cmap, X_colors, X_numcolors);
	    XSync(X_display, False);
	    X_cmapupd = 1;
	    }
	}
#endif
      }
#ifdef XF86DGA
    }
#endif
    break;
#endif
    
  case MappingNotify:
    /* Keyboard remapping.  The book says you can't mask this. */
    if (xevent.xmapping.request == MappingKeyboard ||
	xevent.xmapping.request == MappingModifier) {
      XRefreshKeyboardMapping(&xevent.xmapping);
    }
    break;
    
#if defined(OBSCURED_SLEEP) || defined(FULLSCREEN_AUTORAISE)
  case VisibilityNotify:
#ifdef XF86DGA
    if (!X_dga) {
#endif
      /* Update on visibility on window */
	if (xevent.xvisibility.window == X_mainWindow) {
#ifdef OBSCURED_SLEEP
	  X_visstate = xevent.xvisibility.state;
#endif
#ifdef FULLSCREEN_AUTORAISE      
	  if ((X_dmode & DMODE_FULLSCREEN) &&
	      xevent.xvisibility.state != VisibilityUnobscured) {
	    XRaiseWindow(X_display, X_mainWindow);
	  }
#endif
	}
#ifdef XF86DGA
    }
#endif
    break;
#endif
    
  case SelectionNotify:
    break;
    
  case ClientMessage:
    if (xevent.xclient.message_type == X_atom_WM_PROTOCOLS) {
      if (xevent.xclient.data.l[0] == X_atom_WM_DELETE_WINDOW) {
	/* Tell Acidwarp to quit */
	X_quitflag = 1;
	xpceC_QuitRequest();
      } else {
	  fprintf(stderr, "Unknown WM_PROTOCOLS ClientMessage.\n");
      }
    } else if (xevent.xclient.message_type != ColormapNotify) {
      fprintf(stderr, "Unknown XClientMessage format %u.\n", 
	      xevent.xclient.format);
    }
    break;
    
    /* If we want ConfigureNotify for resizing and enable StructureNotify
	 we also get these unwanted events for free.
	 At least they're rare. */
    
  case MotionNotify:
    X_havemouse = 1;
    X_mousex = xevent.xmotion.x;
    X_mousey = xevent.xmotion.y;
    break;
    
  case CirculateNotify:
  case DestroyNotify:
  case GravityNotify:
  case ReparentNotify:
    break;
    
  case MapNotify:
#ifdef OBSCURED_SLEEP
    X_mwmapped = 1;
#endif
    break;
    
  case UnmapNotify:
#ifdef OBSCURED_SLEEP
    X_mwmapped = 0;
#endif
    break;
    
#ifdef XF86DGA
  case KeyRelease:
    break;
#endif
  default:
#if defined(MITSHM) && defined(DO_SYNC)
    /* XShmPutImage finished */
    if (doShm && xevent.type  == X_shmeventtype) {
      shmFinished = 1;
      break;
    }
#endif
    
    fprintf(stderr, "Unhandled X event: %i\n", (xevent.type));
    break;
  }
}

#ifdef COLORMAP_THEFT
#if 0
/* This function gets the ICCCM compliance level of the window manager */
int x_GetICCCMVer() {
  char *anames[] = { NULL, "INTEGER", "VERSION" };
  Atom atoms[3];

  anames[0] = "WM_S0";
  XInternAtoms(X_display, anames, 3, False, atoms);
  
  fprintf(stderr, "%i, %i, %i\n", atoms[0], atoms[1], atoms[2]);

  /* Send a conversion request */
  XConvertSelection(X_display,
		    atoms[0], /* selection */
		    atoms[1], /* target */ 
		    atoms[2], /* property */ 
		    X_mainWindow,
		    CurrentTime);
}
#endif

int x_XErrorFlaggingHand(Display *d, XErrorEvent *ee) {
  X_error = 1;
  return 0;
}

/* This function causes an ICCCM 2.0 compliant window manager (like
 * WindowMaker) to stop installing colormaps if x == 0 and start
 * installing colormaps if x == 1.  The result is that if you do this
 * you can have OverrideRedirect windows with a colormap, you can
 * run xacidwarp in root with a private colormap, and you can have the
 * acidwarp colormap all the time regardless of focus.
 */
void x_GrabColormap(int x) { 
  XEvent xevent;
  
  /* You're supposed to the ICCCM compliance level of the window manager
   * before you do this but that is not done here because some non 2.0
   * compliant window managers can do this too.
   */
  /* x_GetICCCMVer(); */

  xevent.xclient.type = ClientMessage;
  /* xevent.xclient.serial */
  xevent.xclient.send_event = True;
  xevent.xclient.display = X_display;
  xevent.xclient.window = X_rootwin;
  xevent.xclient.message_type = ColormapNotify; /*WM_COLORMAP_NOTIFY;*/
  xevent.xclient.format = 32;
  xevent.xclient.data.l[0] = CurrentTime;
  xevent.xclient.data.l[1] = x;
  xevent.xclient.data.l[2] = 0;
  xevent.xclient.data.l[3] = 0;
  xevent.xclient.data.l[4] = 0;
  
  XSendEvent(X_display,
	     xevent.xclient.window,
	     False, 
	     ColormapChangeMask,
	     &xevent);

  /* Possible race condition.  There may be pending colormap installations
   * from the window manager which will undo our installation.  However,
   * we are handling ColormapNotify events and that will deal with it.
   */
     
  if (x) {
    XInstallColormap(X_display, X_cmap);
    X_colormapgrabbed |= GRABCMAP_ICCCM;
  } else {
    X_colormapgrabbed &= ~GRABCMAP_ICCCM;
  }
}

void x_GrabRootCMap(int flag) {
  XWindowAttributes     getattribs;
  
  if (flag) {
    if (!(X_colormapgrabbed & GRABCMAP_ROOT)) {
      /* If we haven't grabbed the colormap, save the root window
       * colormap.
       */
      /* Save old root window attributes */
      XGetWindowAttributes(X_display, X_rootwin, &getattribs);
      X_oldrcmap = getattribs.colormap;
    }
    
    /* Set the root window colormap */
    XSetWindowColormap(X_display, X_rootwin, X_cmap);
    X_colormapgrabbed |= GRABCMAP_ROOT;
  } else {
    int                   (*oehandler)(Display *, XErrorEvent *);
      
    XSync(X_display, False);
    oehandler = XSetErrorHandler(x_XErrorFlaggingHand);
    X_error = 0;
    XSetWindowColormap(X_display, X_rootwin, X_oldrcmap);
    XSync(X_display, False);
    XSetErrorHandler(oehandler);
    /* If the XSetWindowColormap call returned an error then that means
     * that the root window had an invalid colormap before.  We'll then
     * just reset it to the default colormap, which is probably right.
     * (Note:  This was mainly an issue during development if the
     * cleanup code didn't get called)
     */
    if (X_error) {
      fprintf(stderr, "Got an error restoring the root window colormap.\n"
	      "Now attempting to set the root window colormap to the default"
	      " colormap.\n");
      XSetWindowColormap(X_display, X_rootwin,
			 DefaultColormap(X_display, X_screen));
    }
    X_colormapgrabbed &= ~GRABCMAP_ROOT;
  }
}
#endif 

void x_CloseWindow() {
  if (X_gc != None) {
    XFreeGC(X_display, X_gc);
    X_gc = None;
  }
  if (X_mainWindow != None && !(X_dmode & DMODE_ROOT)) {
    XDestroyWindow(X_display, X_mainWindow);
    X_mainWindow = None;
  }

#ifdef OBSCURED_SLEEP
  X_visstate = VisibilityFullyObscured;
  X_mwmapped = 0;
#endif
  X_oktodraw = 0;
  X_mainWindow = None;
}


/* xpce_CloseGraph - Cleans up before quitting.  This is only important if
   we have allocated shared memory. */
void xpce_CloseGraph() {
  if (X_xxed) {
    /* It's only possible to get here with the server grabbed if an error
       occurred */
    if (X_servergrabbed) {
      XUngrabServer(X_display);
      X_servergrabbed = 0;
    }

    /* Clean up XFree86-DGA */
#ifdef XF86DGA
    if (X_dga) {
      x_CleanupXF86DGA();
    }
#endif

    /* The undoing of COLORMAP_THEFT goes first because it's the most
       important cleanup */
#ifdef COLORMAP_THEFT
    /* If we were running in the root, restore the window attributes */
    if (X_colormapgrabbed & GRABCMAP_ROOT) {
      x_GrabRootCMap(0);
    }
    
    /* Give colormap control back to the server.  This is important */
    if (X_colormapgrabbed & GRABCMAP_ICCCM) {
      x_GrabColormap(0);
    }

    X_colormapgrabbed = 0;
#endif
    
    /* If we're running in the root window if we don't do something when
     * we quit we will have the ugly effect of our image slowly
     * disappearing because nobody is there to handle the expose events.
     */
    if (X_setbg && X_quitflag) {
      if (X_image != NULL) {
	/* Set root image */
	Pixmap p;
	
	p = XCreatePixmap(X_display, X_rootwin, X_width, X_height, 
			  X_visualinfo->depth);
	x_PutImagePart(p, 0, 0, X_width, X_height);
	XSetWindowBackgroundPixmap(X_display, X_rootwin, p);
	/* The X server creates a copy of the pixmap, and so we can free it */
	XFreePixmap(X_display, p);
	if (X_colormapped) {
	  if (X_privatecmap) {
	    fprintf(stderr,
		    "warning:  The image being used for the background had a "
		    "private colormap.\n          Its colors will be wrong."
		    "\n");
	  } else {
	    /* Dummy pixmap */
	    Pixmap pm = XCreatePixmap(X_display, X_rootwin, 1, 1, 1);
	    
	    if (Xa__XSETROOT_ID == 0) {
	      Xa__XSETROOT_ID = XInternAtom(X_display, 
					    Xan__XSETROOT_ID, False);
	    }

#if 0 /* CLR_REALLOC_RDONLY */
	    /* This is a problem because it may require translation */
	    x_ClrReallocRdonly();
#endif
	    /* Make sure the colormap entries for the image stay safe,
	     * but let them be freed later with xsetroot or a compatible
	     * program.
	     * I figured out how to do this by looking in root.c of 
	     * xloadimage-4.1.
	     */
	    
	    XChangeProperty(X_display, X_rootwin, Xa__XSETROOT_ID, XA_PIXMAP,
			    32, PropModeReplace,
			    (unsigned char *)&pm, 1);
	    
	    /* We need to retain resources. */
	    X_retainres = 1;
	    XSync(X_display, False);
	    fprintf(stderr, "Retaining %i colormap entries for background.  "
		    "Use xsetroot to free them.\n", X_numcolors);
      	  }
	}
      } else {
	fprintf(stderr, "Didn't set background image because there is no "
		"current image\n");
      }
    }
    
    if ((X_setbg && X_quitflag) || (X_dmode & DMODE_ROOT)) {
      XClearWindow(X_display, X_rootwin);
    }

    if (X_windowedx) {
      /* We now start doing lots of cleanup.  This is totally pointless
       * if we just want to terminate the program, but it is needed if we
       * want to change some things.
       */
      
      /* Destroy the XImage.  This is only important if we had shared memory */
      if (X_image != NULL)
	x_DestroyImage();
      
      /* Translator functions go with the image */
      X_sxlatfunc = x_SXlatNull;
      X_xlatdest = NULL;

      X_windowedx = 0;

      /* Destroy the window */
      x_CloseWindow();

      /* Shut down the graphics driver */
      X_vhandler->do_cleanup();
    } /* if (X_windowedx) */

    /* We only do the final stuff if we're really quitting */
    if (X_quitflag) {
      /* Set flag to retain resources if we need that.  Note that this 
       * must only be done here.  Otherwise, resources will get retained
       * even if we haven't done the proper cleanup.
       */
      if (X_retainres) {
	XSetCloseDownMode(X_display, RetainPermanent);
      }

      /* Disconnect from the X server.  Note that if you didn't at least do
       * the XSync here there still may be incoming errors you'll never know
       * about.  It appears that even with the XSync we may not be able
       * to know if we've been killed.  At least since XSetCloseDown mode
       * happens above it shouldn't be a problem.
       */
      XSync(X_display, False);
      XCloseDisplay(X_display);
      fprintf(stderr, "All X cleanup finished.\n");
      
      X_xxed = 0;
    }
  }
}

#ifdef SIGHAND
/* x_SigHand - Handler for signals that should cause the program to quit */
#ifdef HAVE_CONFIG_H
RETSIGTYPE  x_SigHand(int sig)
#else
void x_SigHand(int sig)
#endif
{
  /* Send 'q' to acidwarp */
  X_quitflag = 1;
  xpceC_QuitRequest();
}
#endif

void x_SetNullCursor() {
  Pixmap cursormask;
  XGCValues xgc;
  GC gc;
  XColor dummycolour;
  static Cursor cursor = None;

  /* Create null cursor if necessary */
  if (cursor == None) {
    cursormask = XCreatePixmap(X_display, X_rootwin, 1, 1, 1/*depth*/);
    xgc.function = GXclear;
    gc =  XCreateGC(X_display, cursormask, GCFunction, &xgc);
    XFillRectangle(X_display, cursormask, gc, 0, 0, 1, 1);
    dummycolour.pixel = 0;
    dummycolour.red = 0;
    dummycolour.flags = 04;
    cursor = XCreatePixmapCursor(X_display, cursormask, cursormask,
				 &dummycolour,&dummycolour, 0,0);
    XFreePixmap(X_display,cursormask);
    XFreeGC(X_display,gc);
  }

  XDefineCursor(X_display, X_mainWindow, cursor);
}

#ifdef FREE_OLD_BACKGROUND
void x_FreeOldBackground() {
  Atom type;
  int format;
  unsigned long length, after;
  unsigned char *data;

  /* Get XSETROOT_ID atom */
  if (Xa__XSETROOT_ID == 0) {
    Xa__XSETROOT_ID = XInternAtom(X_display, 
				 Xan__XSETROOT_ID, False);
  } 

  /* This was taken from xsetroot */
  x_GrabServer();  /* Avoid race condition, make sure we kill the client 1st */
  if (XGetWindowProperty(X_display, X_rootwin, Xa__XSETROOT_ID, 0L, 1L, True, 
			 AnyPropertyType,
			 &type, &format, &length, &after, &data) == Success) {
    if ((type == XA_PIXMAP) && (format == 32) &&
	(length == 1) && (after == 0)) {
      XKillClient(X_display, *((Pixmap *)data));
      fprintf(stderr, "Freed old xsetroot resources.\n");
    }
    else if (type != None)
      fprintf(stderr, "warning: _XSETROOT_ID property is garbage\n");
    
    x_UngrabServer();

    XFree(data);
  }
}
#endif

void x_InitXBase() { 
  /* Saved error handler */
  int (*oehandler)(Display *, XErrorEvent *);

  /* open the display */
  X_display = XOpenDisplay(NULL);
  if (!X_display) {
    fprintf( stderr, "Cannot connect to X server \"%s\"\n",
	     XDisplayName(NULL));
    exit(-1);
  }

  /* get the screen (Does :display.screen work?) */
  X_screen = DefaultScreen(X_display);

  /* Detect if a window manager is running.  We do this by attempting to
   * get some events from the root window that can only go to one client
   * and that would be certainly monitored by any reasonable window
   * manager.
   */
  X_rootwin = RootWindow(X_display, X_screen);
  X_error = 0;
  oehandler = XSetErrorHandler(x_XErrorFlaggingHand);
  XSelectInput(X_display, X_rootwin, SubstructureRedirectMask);
  XSync(X_display, False);
  XSetErrorHandler(oehandler);
  XSelectInput(X_display, X_rootwin, 0);
  if (X_error)
    X_wmdetected = 1;

  /* See if we can use MIT-SHM */
#ifdef MITSHM
  if (!X_disableshm) {
    /* check for the MITSHM extension */
    doShm = XShmQueryExtension(X_display);
    
    /* even if it's available, make sure it's a local connection */
    if (doShm)
      {
	if (!X_displayname) X_displayname = (char *) getenv("DISPLAY");
	if (X_displayname)
	  {
	    char *tdn, *d;
            tdn = malloc(strlen(X_displayname)+1);
	    d = tdn; 
	    strcpy (d, X_displayname);
	    while (*d && (*d != ':')) d++;
	    if (*d) *d = 0;
	    if (!(!strcasecmp(d, "unix") || !*d)) doShm = 0;     
	    free(tdn);
	  }
      }
    
    if (doShm)
      fprintf(stderr, "Using MITSHM extension\n");
  }
#endif
    
  X_scr = DefaultScreenOfDisplay(X_display);
  X_screenwidth = WidthOfScreen(X_scr);
  X_screenheight = HeightOfScreen(X_scr);    

  /* The following makes the window come up in the middle of the screen
   * if there is no window manager.
   */
  X_savedx = (X_screenwidth - X_savedwidth) / 2;
  X_savedy = (X_screenheight - X_savedheight) / 2;

  X_xxed = 1;

#ifdef FREE_OLD_BACKGROUND
  if (X_setbg) {
    x_FreeOldBackground();
  }
#endif 
}

void x_InitWindow() {
  unsigned long	        attribmask = 0;
  XSetWindowAttributes  attribs;
  int			valuemask;
  XGCValues		xgcvalues;
  XWMHints              *wmhints;
  XSizeHints            *xsh;
  int                   x, y;
  
  /* setup attributes for main window */
  attribmask |= CWEventMask | CWBorderPixel | CWBackingStore | CWBitGravity
    | CWColormap;

  /* Do not set window gravity here.  That has nothing to do with the 
   * window manager hint later on and it will cause various weirdness
   */

  attribs.event_mask =
    KeyPressMask | ExposureMask | StructureNotifyMask 
#if defined(OBSCURED_SLEEP) || defined(FULLSCREEN_AUTORAISE)
    | VisibilityChangeMask
#endif 
    ;
  attribs.colormap = X_cmap;
  attribs.bit_gravity = ForgetGravity;

  if (X_trackmouse) {
    attribs.event_mask |= PointerMotionMask;
  }
  
#if (defined(COLORMAP_THEFT) && (!defined(MAX_CMAP_REINSTALLS) \
|| MAX_CMAP_REINSTALLS > 0)) || defined(NO_POINTLESS_STORECOLORS)
  if (X_colormapped) {
    attribs.event_mask |= ColormapChangeMask;
  }
#endif

  attribs.border_pixel = 0;
  
  /* If we are using a colormap then backing store is good because
     it avoids resending parts of the image.  However, if we aren't using
     a colormap it is bad because the image changes so often and if 
     backing store is in use the non-visible sections get uselessly
     updated */
  attribs.backing_store = (X_needxlat & NEEDXLAT_FIXCMAP) ?
    NotUseful : WhenMapped;
  
  X_width = X_savedwidth;
  X_height = X_savedheight;

  if (X_dmode & DMODE_ROOT) {
    X_mainWindow = X_rootwin;
    
#ifdef OBSCURED_SLEEP
    /* Assume that it's mapped and visible */
    X_visstate = VisibilityPartiallyObscured;
    X_mwmapped = 1;
#endif
    
    /* The root window background cannot be None.  If it is set to None,
       it will revert to the default.  Therefore XClearArea will always
       repaint the window with some background.  Because of this,
       XClearArea is useless for the root window and we just have to
       blindly put the whole image.
    */
    X_useclear = 0;

#if 1
    /* If we don't set a background pixel we may get some pattern or
     * the default X pattern for the background.  This is pointless if
     * useclear is 0
     */
    attribs.background_pixel = 0;
    XChangeWindowAttributes(X_display, X_rootwin, CWBackPixel,
			    &attribs);
#endif

    /* We must have expose events */
#if (defined(COLORMAP_THEFT) && (!defined(MAX_CMAP_REINSTALLS) \
|| MAX_CMAP_REINSTALLS > 0)) || defined(NO_POINTLESS_STORECOLORS)
    XSelectInput(X_display, X_mainWindow, ExposureMask | 
		 ColormapChangeMask | KeyPressMask | 
		 (X_trackmouse ? PointerMotionMask : 0));

    /* If we don't set the root window to our colormap we won't ever get
       colormap notify events */
#ifdef COLORMAP_THEFT
    if (!(X_grabcmap & GRABCMAP_ROOT))
      X_cmapin = 1;
#endif
#else
    XSelectInput(X_display, X_mainWindow, ExposureMask);
#endif
    
    if (X_privatecmap && X_wmdetected && X_colormapped
#ifdef COLORMAP_THEFT
	&& !X_grabcmap
#endif
	) {
      fprintf(stderr, 
"warning: You are attempting to use a non-default colormap in the root window"
"\n"
"         and a window manager has been detected.  The desired colormap "
"\n"
"         probably won't get installed and all colors will be unchanging and "
"\n"
"         wrong."
#ifdef COLORMAP_THEFT
	        "  Try using the -g or -G options to grab the colormap."
#endif
	      "\n");	     
    }

    X_width = X_screenwidth;
    X_height = X_screenheight;
  } else {
    /* Set full screen size if applicable */
    if (X_dmode & DMODE_FULLSCREEN) {
      X_trackpos = 0;
      X_width = X_screenwidth;
      X_height = X_screenheight;
    } else {
      X_trackpos = 1;
    }

    /* Maybe use OverrideRedirect.  If we are going into fullscreen and
     * it's specified we use it.  But note that we also use it for fullscreen
     * if there's no reason to not use it because of the colormap.
     */
    if ((X_dmode & DMODE_FULLSCREEN) && ((X_fsmode & FSMODE_OR) ||
	!X_colormapped || !X_privatecmap)) {
      attribmask |= CWOverrideRedirect; 
      attribs.override_redirect = True;
    }

    /* Create the main window */
    if (X_dmode & DMODE_FULLSCREEN) {
      x = 0;
      y = 0;
    } else {
      x = X_savedx;
      y = X_savedy;
    }
    X_useclear = 1;   /* We can use XClearArea to get exposes in our window */
    X_mainWindow = XCreateWindow(X_display,
				 X_rootwin,
				 x, y,
				 X_width, X_height,
				 0, /* borderwidth */
				 X_visualinfo->depth,
				 InputOutput,
				 X_visual,
				 attribmask,
				 &attribs);

    if (X_dmode & DMODE_FULLSCREEN) {
      x_SetNullCursor();
    }

    /* Set window title */
    XStoreName(X_display, X_mainWindow, X_caption);

    /* Setup window manager hints */
    wmhints = XAllocWMHints();
    wmhints->flags = InputHint | StateHint;
    /* If this input hint isn't there then with twm the window needs
       to lose and regain focus before it starts getting keyboard input */
    wmhints->input = True;
    wmhints->initial_state = NormalState;
    XSetWMHints(X_display, X_mainWindow, wmhints);
    XFree(wmhints);

    /* If in "full screen using window manager hints" mode then setup
       size hints */    
    xsh = XAllocSizeHints();

    xsh->flags = PWinGravity;
    xsh->win_gravity = StaticGravity;

    if ((X_dmode & (DMODE_FULLSCREEN | DMODE_XSET | DMODE_YSET)) ||
	X_psrem) {
      xsh->flags |= USSize;
      xsh->width = X_width; xsh->height = X_height;
    }

    if (X_dmode & DMODE_FULLSCREEN) {
      /* Important:  This is a user specified position and size.  (The user
       * set the command line option.)  This makes more window managers
       * respect the hint.
       */
      xsh->x = 0; xsh->y = 0;
      xsh->flags |= USPosition;
    } else if (X_psrem) {
      xsh->x = X_savedx; xsh->y = X_savedy;
      xsh->flags |= USPosition;
    }

    XSetWMNormalHints(X_display, X_mainWindow, xsh);

    XFree(xsh);

    /* Setup window manager protocols */
    if (X_atom_WM_PROTOCOLS == 0) {
      XInternAtoms(X_display, X_wmpatomnames, 2, False, &X_wmpatoms[0]);
    }
    XChangeProperty(X_display, X_mainWindow, X_atom_WM_PROTOCOLS,
		    XA_ATOM, 32, PropModeReplace,
		    (unsigned char *)&X_atom_WM_DELETE_WINDOW, 1);

    /* Note:  We don't need WM_COLORMAP_WINDOWS.  We would only need that
     *        if we had some child windows that needed colormap that differed
     *        from our main window.
     */
  }
  
  /* create the GC */
  valuemask = GCGraphicsExposures;
  xgcvalues.graphics_exposures = False;
  X_gc = XCreateGC(X_display, X_mainWindow, valuemask, &xgcvalues);
  
#ifdef COLORMAP_THEFT
  if (X_privatecmap && X_colormapped && X_grabcmap &&
      (!X_grabcmapfsr || X_dmode & (DMODE_ROOT | DMODE_FULLSCREEN))) {
    if (X_grabcmap & GRABCMAP_ICCCM)
      x_GrabColormap(1);
    if (X_grabcmap & GRABCMAP_ROOT)
      x_GrabRootCMap(1);
  }
#endif
  
  /* map the window */
  if (!(X_dmode & DMODE_ROOT)) {
    XMapWindow(X_display, X_mainWindow);
    
    if (X_dmode & DMODE_FULLSCREEN) {
      XRaiseWindow(X_display, X_mainWindow);
      
      if ((X_fsmode & FSMODE_OR) ||
	  !X_colormapped || !X_privatecmap) {
	XSetInputFocus(X_display, X_mainWindow, RevertToPointerRoot, 
		       CurrentTime);
      }
    }
    
    /* wait until it is OK to draw */
    if (!X_useclear) {
      while (!X_oktodraw) {
	x_GetEvent();
      }
    }
  }
}

void xpce_InitGraph() {
  /* Init some variables */
  X_quitflag = 0;
  X_havemouse = 0;

  /* Check for incompatible command line options */
  if (X_setbg && !(X_dmode & (DMODE_ROOT | DMODE_FULLSCREEN)))
    x_Error("Background setting is only possible with root or full screen"
	    " modes.\n");

  /* Initialize the base X stuff that's always needed if we want to use X */
  if (!X_xxed)
    x_InitXBase();
    
  /* The root window has the default visual.  We can't display non-default
     visual stuff in there */
  if (X_dmode & DMODE_ROOT) {
    if ((X_visdeftype & VISDEF_ID) && 
	X_visdef_id != XVisualIDFromVisual(DefaultVisual(X_display, 
							 X_screen))) {
      x_Error("To display in the root window you need to use the default "
	      "visual.\n");
    }
    X_visdeftype |= VISDEF_DEFAULT;
  }
      
  /* Select a visual handler and visual class */
  X_needxlat = 0;
  X_xalloccolor &= XALLOCCOLOR_USER;
#ifdef XF86DGA
  if ((X_dmode & DMODE_FULLSCREEN) && (X_fsmode & FSMODE_DGA)) {
    x_InitXF86DGA();
    X_windowedx = 0;
    X_vhandler->do_initxlat();

    return;
  }
  else
#endif
    x_PickVisual();

  /* If it's a special visual class then don't do more stuff */
  if (X_vhandler->flags & VISINFO_NOTXWIN) {       
    X_windowedx = 0;
    X_vhandler->do_setup(); 
    X_vhandler->do_initxlat();
    return;
  }
  
  /* We will now set up things in the usual X way */
  X_windowedx = 1;

  fprintf(stderr, "Using %s visual handler with depth = %i.\n",
	  X_vhandler->name, X_visualinfo->depth);

  /* Call the setup function for the visual. */
  X_vhandler->do_setup(); 

  x_InitWindow();
  
  x_CreateImage();

  /* We now have the red, green and blue masks if needed, so we can 
   * finally setup the mapping 
   */
  X_vhandler->do_initxlat();
}

/* x_PutImage - Actually update the image on the X server from our
 *              XImage structure.
 */
inline void x_PutImage() {
#ifdef XF86DGA
  if (!X_dga) {
#endif
    if (X_useclear) {
      XClearArea(X_display, X_mainWindow, 0, 0, 0, 0, True);
    } else {
      x_PutImagePart(X_mainWindow, 0, 0, X_width, X_height);
    }
#ifdef XF86DGA
  }
#endif
}

/******************* ACIDWARP INTERFACE FUNCTIONS ****************************/

/* xpce_DrawImage - Do whatever needs to be done when acidwarp wants to 
 *                  draw a new image.  We first translate the image if 
 *                  needed.  Then we upload it to the server.
 */
void xpce_DrawImage() {
#if defined(XF86DGA) && XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER 
  if (X_dga && X_dblbuf)
    x_XF86DGASwapBuf();
#endif

  if (X_needxlat) { 
#ifdef TIME_XLAT
    int x;
    #include <sys/times.h>
    struct tms before, after;
    times(&before);
    for (x = 0; x < TIME_XLAT; x++) {
#endif
      x_XlatBuffer();
#ifdef TIME_XLAT
    }
    times(&after);
    fprintf(stderr, "%i * XlatTime = %i\n", 
	    TIME_XLAT,
	    after.tms_utime - before.tms_utime);
#endif
  }
  x_PutImage();
}

/* xpce_SetPalette - Do whatever needs to be done when acidwarp sets a new 
 *                   palette.  If we are using a colormap then that means
 *                   uploading the new colormap to the server.  If we
 *                   aren't using a new colormap this means redrawing the
 *                   whole image.
 */
void xpce_SetPalette (unsigned char* palette)
{
  /* Update translation table */
  X_vhandler->do_updatexlat(palette);

#ifndef NO_PAL_REDRAW  
  /* The image must be redrawn if we aren't in colormapped mode. */
  if (X_needxlat & NEEDXLAT_ON_CMAPCHG)
    xpce_DrawImage();
#endif
}

/* xpce_HandleEvents - Handle all pending X events */
void xpce_HandleEvents (void)
{
  /* This function has been made somewhat less compact and efficient
   * for easier debugging.  Who cares; it's not like this code gets
   * executed a lot.
   */
  int anyways, numpending;

    while (1) {
      anyways = 0;
#ifdef OBSCURED_SLEEP
	 /* Note:  If you minimize the window in WindowMaker we just get
	  *        an UnmapNotify and no VisibilityNotify, and so we also
          *        need to keep track of the mapped state of the main
	  *        window.
	  */
      if ((X_visstate == VisibilityFullyObscured || !X_mwmapped)
	  && !X_quitflag
#ifdef COLORMAP_THEFT
	  && !X_colormapgrabbed 
#endif
#ifdef XF86DGA
	  && !X_dga
#endif
	  ) anyways = 1;
#endif

      if (!anyways) {
	/* fprintf(stderr, "XPending->"); */
	numpending = XPending(X_display);
	/* fprintf(stderr, "<-XPending"); */
	if (numpending == 0) break;
      }
      
      x_GetEvent();
    }
  /* Make sure that the display gets updated before we go back */
  XFlush(X_display);
}

unsigned int xpce_GetHeight() {
  return X_height;
}

unsigned int xpce_GetWidth() {
  return X_width;
}

/* x_GetNumColors - returns number of colors available to Acidwarp. */
unsigned int xpce_GetNumColors() {
  /* The number apparently only has to be greater than 3.  However, if it's
     not 256 then various modes have defects, and so we're lying here. */
  return INPAL_SIZE;
}

void xpce_QuitNotify() {
  X_quitflag = 1;
}

/* FIXME this does a pointless memcpy at 8bpp */
void xpce_BlitAll(unsigned char *b) {
#ifndef NO_BUFGRAF
  unsigned char *oldbg = buf_graf;
#endif
  buf_graf = b;
  if (!X_needxlat) {
    memcpy(X_xlatdest, b, X_height * X_width);
  }
  xpce_DrawImage();
#ifndef NO_BUFGRAF
  buf_graf = oldbg;
#endif
}

/********************** COMMAND LINE FUNCTIONS *******************************/

/* xpce_CLHelp - Prints out help on the command line options */
void xpce_CLHelp() {
  printf("\n"
"The X Window display support has many options.  You probably won't need any\n"
"of them if you just want to display Acidwarp in a normal X Window.\n"  
"\n"
"'R' - Display in the root window\n"
"'B' - Retain last image as root window background\n"
"'f' - Display in full-screen mode\n"
"'F' - Use OverrideRedirect for full screen mode\n"
#ifdef XF86DGA
"'D' - Use XFree86-DGA for full screen mode\n"
#endif
"'x' - Set width.  Note that you can also resize the window.\n"
"'y' - Set height.  Note that you can also resize the window.\n"
"'v' - Select a visual handler; give ? as an argument to list handlers\n"
"'W' - Select a visual handler for full screen mode\n"
"'V' - Select a visual class; use Default or an id from xdpyinfo\n"
"'I' - Use a private colormap\n"
"'i' - Use a shared colormap\n"
"'c' - Set the number of colors\n"
#ifdef COLORMAP_THEFT
"'g' - Grab control of the colormap using ICCCM 2.0\n"
"'G' - Grab the colormap by making ours the root window colormap\n"
"'P' - Only grab colormap when in root or full screen mode\n"
#endif
#ifdef MITSHM
"'S' - Disable use of MIT-SHM extension\n"
#endif
"'A' - Always use XAllocColor to get pixel values (slow!)\n"
"'P' - Use XPutPixel to draw into XImage (slow!)\n"
);
}

void x_SetDmode(int dmode, int arg) {
  int t;
  
  t = X_dmode & ~(DMODE_XSET | DMODE_YSET);
  if (t != DMODE_UNSPECIFIED && t != dmode) {
    x_Error("Conflicting display mode options.\n");
  }
  X_dmode |= dmode;
}
     
char *x_GetOptArg(int argc, char **argv, int *argNum) {
  if((argc-1) > *argNum) {
    (*argNum)++;
    return argv[*argNum];
  } else {
    fprintf(stderr, "No argument for option '%c'\n", argv[*argNum][1]);
    exit(-1);
  }
}

long int x_GetOptIntArg(int argc, char **argv, int *argNum, int base) {
  char *asptr, *endptr;
  long int r;
  char o;

  o = argv[*argNum][1];
  asptr = x_GetOptArg(argc, argv, argNum);

  if (*asptr == 0) goto goiaerr;

  r = strtol(asptr, &endptr, base);
  if (r == LONG_MIN || r == LONG_MAX || *endptr != 0) {
  goiaerr:
    fprintf(stderr, "Invalid integer argument for option '%c'\n", o);
    exit(-1);
  }
  
  return r;
}

void x_SetCmapPref(int i) {
  if (X_cmappref == CMAPPREF_UNSPEC) {
    X_cmappref = i;
  } else if (X_cmappref != i) {
    x_Error("Conflicting colormap preferences\n");
  }
}

int xpce_ProcessOption(int argc, char **argv, int *argNum) {
  char *opt = argv[*argNum];
  char *arg;
  int intarg;

  if (strlen(opt) == 2 && opt[0] == '-') {
    switch (opt[1]) {
    case 'R':
      x_SetDmode(DMODE_ROOT, 0);
      break;

    case 'F':
      if (X_fsmode != 0 && X_fsmode != FSMODE_OR) {
	x_Error("Conflicting full screen modes specified.\n");
      }
      X_fsmode = FSMODE_OR;
      break;

#ifdef XF86DGA
    case 'D':
      if (X_fsmode != 0 && X_fsmode != FSMODE_DGA) {
	x_Error("Conflicting full screen modes specified.\n");
      }
      X_fsmode = FSMODE_DGA;
      break;
#endif

    case 'f':
      x_SetDmode(DMODE_FULLSCREEN, 0);
      break;

    case 'i':
      x_SetCmapPref(CMAPPREF_DEFAULT);
      break;

    case 'I':
      x_SetCmapPref(CMAPPREF_PRIVATE);
      break;

#ifdef COLORMAP_THEFT
    case 'g':
      X_grabcmap |= GRABCMAP_ICCCM;
      break;

    case 'G':
      X_grabcmap |= GRABCMAP_ROOT;
      break;

    case 'p':
      X_grabcmapfsr = 1;
      break;
#endif

    case 'c':
      X_u_numcolors = x_GetOptIntArg(argc, argv, argNum, 10);
      if (X_u_numcolors < 2 || X_u_numcolors > INPAL_SIZE) {
	fprintf(stderr,"Invalid number of colors selected (range is"
		" 2 to %i)\n", INPAL_SIZE);
	exit(-1);
      }
      break;

    case 'x':
      intarg = x_GetOptIntArg(argc, argv, argNum, 10);
      if ((X_dmode & DMODE_XSET) && (X_savedx != intarg)) {
        x_Error("Conflicting horizontal sizes specified.");
      }
      if (intarg < 1) {
	x_Error("Specified width must be greater than zero.\n");
      }
      X_dmode |= DMODE_XSET;
      X_savedwidth = intarg;      
      break;

    case 'y':
      intarg = x_GetOptIntArg(argc, argv, argNum, 10);
      if ((X_dmode & DMODE_YSET) && (X_savedy != intarg)) {
        x_Error("Conflicting vertical sizes specified.");
      }
      if (intarg < 1) {
	x_Error("Specified height must be greater than zero.\n");
      }
      X_dmode |= DMODE_YSET;
      X_savedheight = intarg;
      break;
      
    case 'v':
      arg = x_GetOptArg(argc, argv, argNum);
      X_visdef_handler = x_SetWantedVis(arg, VISDEF_HANDLER);	
      break;
      
    case 'V':
      arg = x_GetOptArg(argc, argv, argNum);
      if (X_visdeftype & (VISDEF_DEFAULT | VISDEF_ID)) {
	x_Error("The -V option may only be used once\n");
      } else {
	if (!strcmp("Default", arg)) {
	  X_visdeftype |= VISDEF_DEFAULT;
	} else {
	  char *endptr;

	  if (arg[0] == 0) goto opt_V_err;
	  X_visdef_id = strtol(arg, &endptr, 0);
	  if (*endptr != 0) {
	  opt_V_err:
	    x_Error("Invalid visual class ID for -V option.\n");
	  }
	  X_visdeftype |= VISDEF_ID;
	}
      }
      break;

    case 'W':
      arg = x_GetOptArg(argc, argv, argNum);
      X_visdef_fshand = x_SetWantedVis(arg, VISDEF_FSHAND);	
      break;

    case 'A':
      X_xalloccolor |= XALLOCCOLOR_USER;
      break;

#ifdef MITSHM
    case 'S':
      X_disableshm = 1;
      break;
#endif

    case 'B':
      X_setbg = 1;
      break;

    case 'P':
      X_xputpixel = 1;
      break;

    default:
      return 0;  /* Unrecognized option */
    }
    return 1;  /* Option was recognized */
  }
  return 0;  /* Argument not in option format */
}

/* NEW for G-Force */
void xpce_SetWindowCaption(char *s) {
  X_caption = s;
  if (X_mainWindow) {
    XStoreName(X_display, X_mainWindow, X_caption);
  }
}

void xpce_QueryMouse(int *x, int *y) {
  if (X_havemouse) {
    *x = X_mousex;
    *y = X_mousey;
  } else if (X_mainWindow != None) {
    int _x, _y, m;
    Window r, c;
    XQueryPointer(X_display, X_mainWindow, &r, &c, &_x, &_y, x, y, &m);
  }
}

  /* The atexit crashes XMMS */
void xpce_InstallErrHand() {
  /* Install our error handlers if we haven't done that before */
  if (!X_ehinst) {
#ifdef XF86DGA
    /* The XF86DGA fork'd better be near the beginning to avoid the 
     * possibility that this program is already far too fucked up 
     * to rescue itself.
     */
    if (X_fsmode & FSMODE_DGA) {
      XF86DGAForkApp(X_screen);
    }
#endif
    X_ehinst = 1;
    atexit(xpce_CloseGraph);
#ifdef SIGHAND
    SETSIGHAND
#endif
  }
}

