/* xvhandler.c - Video drivers for the LibXPCE, the X PseudoColor emulation
 *               library.  This file contains functions that set up and use
 *               various visual classes.
 *
 * Copyright (C) 2000 Boris Gjenero
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1, as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You may contact Boris Gjenero at bgjenero@sympatico.ca
 */

#include "xconfig.h"

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#ifdef XF86DGA
#include <unistd.h>   /* for setuid, getuid */
#include <X11/extensions/xf86dga.h>
#endif
#include "x.h"

/* The visual that the user wants */
int X_visdef_handler;

/* The full screen visual handler that the user wants */
int X_visdef_fshand;

/* The visual class ID that the user wants */
int X_visdef_id;

/* Which one of the above things are defined */
int X_visdeftype = VISDEF_UNDEFINED;

/* Info about the visual we'll use */
visinfo_t *X_vhandler;

/* Bits that are set outside the r,g,b masks (can this ever happen?) */ 
unsigned long int X_mapbase = 0;

/* Maps for each color mapping intensity to component value */
#define X_graymap X_redmap
#if GRAYXLAT_SIZE > INPAL_CVALUES
unsigned long int X_redmap[GRAYXLAT_SIZE];
#else
unsigned long int X_redmap[INPAL_CVALUES];
#endif
unsigned long int X_greenmap[INPAL_CVALUES];
unsigned long int X_bluemap[INPAL_CVALUES];
/* Masks to go along with the maps */
unsigned long X_red_mask = 0, X_green_mask = 0, X_blue_mask = 0;

#ifdef SCXLAT
/* Variables used by the StaticColor XAllocColor replacement */
XColor *X_cmapinfo = NULL;
int X_cmapinfoent = 0;
#endif

#ifdef XF86DGA
int X_dga = 0;
unsigned char *X_dga_addr;

#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
/* Double buffering enabled */
int X_dblbuf;

/* Current page when double buffering */
int X_dblbufbuf;
#endif
#endif

/*********************** VISUAL HANDLER STUFF *******************************/

/* Default visual setup function that does nothing */
void x_NullFunction() {
}

void x_NoWMCmapInst() {
  if (!X_wmdetected) {
    fprintf(stderr, "No window manager was detected and we are using a "
	    "private colormap,\nso the colormap will be installed.\n");
#ifdef NO_POINTLESS_STORECOLORS
    X_cmapin = 1;
#endif
    XInstallColormap(X_display, X_cmap);
  }
}

/* x_FreeColors - free all our colors */
void x_FreeColors() {
  unsigned long pixels[INPAL_SIZE];
  int i;

  if (X_numcolors > 0) {
    /* Fill out list of pixels so we can free them all at once */
    for (i = 0; i < X_numcolors; i++) {
      pixels[i] = X_colors[i].pixel;
    }
    
    /* Free the colors */
    XFreeColors(X_display, X_cmap, &(pixels[0]), X_numcolors, 0);
    X_numcolors = 0;
  }
}

#ifdef CLR_REALLOC_RDONLY
void x_ClrReallocRdonly() {
  int i;

  /* We don't want anyone else to take our colors before we can reallocate
     them */
  x_GrabServer();

  /* Free all our colors */
  x_FreeColors();

  /* Reallocate the colors read-only */
  for (i = 0; i < X_numcolors; i++) {
    if (!XAllocColor(X_display, X_cmap, &(X_colors[i]))) {
      x_UngrabServer();
      x_Error("Failed to reallocate color.\n");
    }    
  }
  
  x_UngrabServer();
}
#endif

/* x_InitColormapped - Initially sets up the colormap */
void x_InitColormapped() {
  int i;
  
  X_privatecmap = PRIVATECMAP;

  /* If we aren't using the default visual then there is no such thing 
     as a default colormap. */
  if (XVisualIDFromVisual(X_visual) !=
      XVisualIDFromVisual(DefaultVisual(X_display, X_screen)) && 
      X_colormapped) {
    if (X_cmappref == CMAPPREF_DEFAULT) {
      x_Error("Cannot use the default colormap with a non-default"
	      " visual.\n");
    } else {
      X_privatecmap = 1;
    }
  }
  
  /* If the user didn't give any preference about the colormap then maybe
     decide here */
  if (X_cmappref == CMAPPREF_UNSPEC) {
    if ((X_dmode & 
	 (DMODE_ROOT | ((X_fsmode & FSMODE_OR) ? DMODE_FULLSCREEN : 0))) && 
#ifdef COLORMAP_THEFT
	!X_grabcmap &&
#endif
	X_wmdetected) {
      /* Under these conditions we probably can't install our own colormap,
	 so we'll try to use the default one */
      X_privatecmap = 0;
    }
  } else if (X_cmappref == CMAPPREF_PRIVATE) {
    X_privatecmap = 1;
  } else if (X_cmappref == CMAPPREF_DEFAULT) {
    X_privatecmap = 0;
  }

  if (!X_privatecmap) {
    /* Use the default colormap */
    X_cmap = DefaultColormap(X_display, X_screen);
#ifdef NO_POINTLESS_STORECOLORS
    /* I don't know if there's any safe way to see if it's installed now,
       so... */
    X_cmapin = 1;
#endif
    
    /* Keep allocating colors until we reach the maximum we want or an
       allocation fails */
    for (i = 0; i < X_u_numcolors; i++) {
      if (!XAllocColorCells(X_display, X_cmap, 0 /* non-contig */,
			    NULL, 0, &X_colors[i].pixel, 1)) {
	/* Allocation failure.  Update number of colors 
	   and stop allocating */
	break;
      }
      X_colors[i].flags = DoRed|DoGreen|DoBlue;
    }
    X_numcolors = i;

    if (i == 0) {
      x_Error("Couldn't allocate any colormap entries.\n");
    }
    
    fprintf(stderr, "Using %i entries in standard colormap.\n", 
	    X_numcolors);

  } else { /* Private colormap in colormapped mode */
    /* Limit us to the colormap size */
    if (INPAL_SIZE > X_visualinfo->colormap_size) {
      X_numcolors = X_visualinfo->colormap_size;
    } else {
      X_numcolors = INPAL_SIZE;
    }

    if (X_numcolors == 0) {
      x_Error("Colormap has no entries.\n");
    }
    
    if (X_visualinfo->class != DirectColor) {
      /* Create our colormap */
      X_cmap = XCreateColormap(X_display, 
			       X_rootwin,
			       X_visual, AllocAll);
      
      /* Initialize the colormap.  Note that it is okay to fill pixel values
	 like this.  Look at the XCreateColormap man page. */
      for (i = 0; i < X_numcolors; i++) {
	X_colors[i].pixel = i;
	X_colors[i].flags = DoRed|DoGreen|DoBlue;
	X_colors[i].red = X_colors[i].green = X_colors[i].blue = 0;
      }
    } else {
      /* Create our colormap */
      X_cmap = XCreateColormap(X_display, 
			       X_rootwin,
			       X_visual, AllocNone);
      
      /* In DirectColor we use XAllocColorcells so we don't have to worry
	 about creating pixel values that don't conflict with each other */
      for (i = 0; i < X_visualinfo->colormap_size; i++) {
	if (!XAllocColorCells(X_display, X_cmap, 0 /* non-contig */,
			      NULL, 0, &X_colors[i].pixel, 1)) {
	  /* Allocation failure.  Update number of colors 
	     and stop allocating */
	  X_numcolors = i;
	  break;
	}
	X_colors[i].flags = DoRed|DoGreen|DoBlue;
      }
    }

    fprintf(stderr, "Using %i colors in private colormap.\n", X_numcolors);
    
    /* Install colormap if no window manager was detected */
    x_NoWMCmapInst();
  }
    
  /* Is the colormap fucked up so we need translation? */
  if (X_colors[0].pixel != 0 || X_numcolors < INPAL_SIZE) {
    X_needxlat |= NEEDXLAT_CMAP;
  } else {
    for (i = 1; i < X_numcolors; i++) {
      if (X_colors[i].pixel != X_colors[i-1].pixel+1) {
	X_needxlat |= NEEDXLAT_CMAP;
	break;
      }
    }
  }
}

/* Initialize in the fixed colormap case */
void x_InitFixCMap() {
  if (XVisualIDFromVisual(X_visual) ==
      XVisualIDFromVisual(DefaultVisual(X_display, X_screen))) {
    /* We aren't colormapped, and we are using the default visual, so
       we can just use the default colormap */
    X_cmap = DefaultColormap(X_display, X_screen);
  } else {
    /* We aren't colormapped and we are using a non-default visual, so
       we create a colormap just for a lookup table.  Weird but
       necessary. */
    X_cmap = XCreateColormap(X_display, 
			     X_rootwin,
			     X_visual, AllocNone);
    X_privatecmap = 1;
  }

  /* With a fixed colormap we definitely need translation */
  X_needxlat |= NEEDXLAT_FIXCMAP;

#ifndef NO_PAL_REDRAW  
  fprintf(stderr, "Emulating a colormap by redrawing the image.\n");
#endif
}

/* x_InitCmapXlat - Initialize translation table in colormapped cases */
void x_InitCmapXlat() {
  /* This will be the translation table used when initially putting the
   * display on the screen.  Unlike in the non-colormapped case,
   * it stays constant.
   */
  
  if (X_numcolors == INPAL_SIZE) {
    int i;
    for (i = 0; i < INPAL_SIZE; i++) {
      X_xlattab[i] = X_colors[i].pixel;
    }
  } else {
    /* We don't have all the palette entries we want */
    int istart, i, j, t;
    int cend;
    
    /* In Acidwarp, 0 is always black, and we should probably leave that
     * alone.
     */
    if (X_numcolors > 1) {
      istart = 1;
      j = 1;
      X_xlattab[0] = X_colors[0].pixel;
    } else {   
      istart = 0;
      j = 0;
    }
    for (i = istart; i < X_numcolors; i++) {
    /* Figure out end of chunk */
      cend = istart + ((i + 1 - istart) * 
		       (INPAL_SIZE - istart) / (X_numcolors - istart));
      
      t = X_colors[i].pixel; 
      for (; j < cend; j++) {
	X_xlattab[j] = t;
      }
    }
  }

  /* Fix endianness if necessary */
  x_Endiannize(&X_xlattab[0], sizeof(unsigned long int), INPAL_SIZE);

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

/* x_DCTCEDoComp() - Do one component for DCTCE */
int x_DCTCEDoComp(int mask, int color) {
  static const char * cnames[] = { "red", "green", "blue" };
  static const char cflags[] = { DoRed, DoGreen, DoBlue }; 
  static unsigned long int * const carrays[] = { X_redmap, 
						 X_greenmap, 
						 X_bluemap };
  unsigned long int planes[INPAL_CBITS];
  int numplanes = 0;
  int i, j, t;
  XColor c;

  if (mask == 0) {
    fprintf(stderr, "No planes in %s component in TrueColor emulation on "
	    "DirectColor\n", &cnames[color]);
    exit(-1);
  }

  /* Setup planes array */
  t = 1;
  while (mask != 0) {
    if (mask & 1) {
      planes[numplanes++] = t;
    }

    mask >>= 1;
    t <<= 1;
  }

  /* Setup everything */
  for (i = 0; i < (1 << numplanes); i++) {
    /* Build up pixel */
    c.pixel = X_mapbase;
    for (j = 0; j < numplanes; j++) {
      if (i & (1 << j)) {
	c.pixel |= planes[j];
      }
    }

    /* Set up palette value */
    c.red = (i << (16 - numplanes));
    c.green = c.red;
    c.blue = c.red;
    c.flags = cflags[color];
    XStoreColor(X_display, X_cmap, &c);

    /* Fill out band of color lookup table entries */
    for (j = 0; j < (1 << (6 - numplanes)); j++) {
      carrays[color][(i << (6 - numplanes)) + j] = c.pixel;
    }
  }
  
  return numplanes;
}


/* x_InitDCTCE() - Initializer for TrueColor emulation in DirectColor mode */
void x_InitDCTCE() {
  int rbits, gbits, bbits, i;
  unsigned long int rmask, gmask, bmask;

  if (X_cmappref == CMAPPREF_DEFAULT) {
    X_cmap = DefaultColormap(X_display, X_screen);

    /* This will allocate the maximum number of color planes we can get
     * off the standard colormap.  Note that actually, there could be
     * a small number of one color and lots of other colors.  In that case
     * we'd get the minimum.  Oh well.
     */
    for (i = INPAL_CBITS; ; i--) {
      if (i == 0) {
	x_Error("No color planes could be obtained.");
      }
      if (XAllocColorPlanes(X_display, X_cmap, False, &X_mapbase, 1, 
			    i, i, i,
			    &rmask, &gmask, &bmask)) {
	break;
      }
    }
  } else {
    X_privatecmap = 1;

    /* Setup private colormap */
    X_cmap = XCreateColormap(X_display, 
			     X_rootwin,
			     X_visual, AllocAll);

    /* Install colormap if no window manager was detected */
    x_NoWMCmapInst();
    
    /* This must be 0.  See the XCreateColormap man page */
    X_mapbase = 0;

    /* Yay, we own all the planes.  See the man page if you want proof. */
    rmask = X_visualinfo->red_mask;
    gmask = X_visualinfo->green_mask;
    bmask = X_visualinfo->blue_mask;
  }

  /* Setup the actual colors and the lookup tables */
  rbits = x_DCTCEDoComp(rmask, 0);
  gbits = x_DCTCEDoComp(gmask, 1);
  bbits = x_DCTCEDoComp(bmask, 2);

  fprintf(stderr, "Using TrueColor emulation on DirectColor, "
	  "R:G:B = %i:%i:%i\n", rbits, gbits, bbits);

  X_needxlat |= NEEDXLAT_FIXCMAP;
}

/* x_InitTrueColorXlat - Initialize translation table for TrueColor 
 *                       This avoids the use of XAllocColor later
 *                       on.  This speeds things up and it should be compatible
 *                       with any hardware.
 */
void x_InitTrueColorXlat() {
  int i;
  XColor c;
  int basecand;
  
  if (X_xalloccolor)
    return;

  /* We use XAllocColor to get the values of the components.  
   * red, green and blue are set to the same value, so that we
   * look up all three at once.  This should be okay because
   * they are independent.
   */
  c.flags = DoRed | DoGreen | DoBlue;
  X_mapbase = 0;
  for (i = 0; i < INPAL_CVALUES; i++) {
    c.red = i << INPAL_SHIFT;
    c.green = c.red;
    c.blue = c.red;

    if (!XAllocColor(X_display, X_cmap, &c)) {
      /* This should *never* happen */
      x_Error("XAllocColor failed in a non-colormapped visual\n");
    }
    
    /* Store the mask values */
    X_redmap[i] = c.pixel & X_red_mask;
    X_greenmap[i] = c.pixel & X_green_mask;
    X_bluemap[i] = c.pixel & X_blue_mask;
    basecand = c.pixel & 
      ~(X_red_mask | X_green_mask | X_blue_mask);
    if (i > 0) {
      if (X_mapbase != basecand) {
	fprintf(stderr, "Messed up pixel masks.  XAllocColor will be used "
		"instead.\n");
	X_xalloccolor |= XALLOCCOLOR_PROGRAM;
	return;
      }
    } else {
      X_mapbase = basecand;
    }
  
    /* We don't actually need the color. */
    XFreeColors(X_display, X_cmap, &c.pixel, 1, 0);
    /* Hopefully the X server doesn't care that we will use non-allocated
     * colors.  It certainly shouldn't care because in TrueColor the mapping
     * is supposed to be static.  
     */
  }


  /* We will fix endianness in the maps for each color now, so that
   * we don't need to fix endianness in X_xlattab always.
   */
  x_Endiannize(&X_mapbase, sizeof(unsigned long int), 1);
  x_Endiannize(&(X_redmap[0]), sizeof(unsigned long int), INPAL_SIZE);
  x_Endiannize(&(X_greenmap[0]), sizeof(unsigned long int), INPAL_SIZE);
  x_Endiannize(&(X_bluemap[0]), sizeof(unsigned long int), INPAL_SIZE);
     
  /* Initialize the colormap to all zeros */
  for (i = 0; i < INPAL_SIZE; i++) {
    X_xlattab[i] = X_mapbase | 
      X_redmap[0] | X_greenmap[0] | X_bluemap[0];
  }

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

/* x_InitStaticGrayXlat - Initialize translation table for TrueColor 
 *                        This avoids the use of XAllocColor later
 *                        on.  This speeds things up and it should be
 *                        compatible with any hardware.  I don't know if
 *                        XAllocColor maybe maps color to grayscales in a 
 *                        better way.
 */
void x_InitStaticGrayXlat() {
  int i;
  XColor c;
  
  if (X_xalloccolor)
    return;

  /* We use XAllocColor to get the values of the components.  
   * red, green and blue are set to the same value, so that we
   * look up all three at once.  This should be okay because
   * they are independent.
   */
  c.flags = DoRed | DoGreen | DoBlue;
  X_mapbase = 0;
  for (i = 0; i < GRAYXLAT_SIZE; i++) {
    c.red = i << (16 - GRAYXLAT_BITS);
    c.green = c.red;
    c.blue = c.red;

    if (!XAllocColor(X_display, X_cmap, &c)) {
      /* This should *never* happen */
      x_Error("XAllocColor failed in a non-colormapped visual\n");
    }
    
    /* Store the mask values */
    X_graymap[i] = c.pixel;

    /* We don't actually need the color. */
    XFreeColors(X_display, X_cmap, &c.pixel, 1, 0);
    /* Hopefully the X server doesn't care that we will use non-allocated
     * colors.  It certainly shouldn't care because in TrueColor the mapping
     * is supposed to be static.  
     */
  }

  /* We will endiannize the graymap so that we don't need to endiannize
   * the translation table every time we change it.
   */
  x_Endiannize(&(X_graymap[0]), sizeof(unsigned long int), GRAYXLAT_SIZE);


  for (i = 0; i < INPAL_SIZE; i++) {
    X_xlattab[i] = X_graymap[0];
  }

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

#ifdef SCXLAT
/* x_InitSCXlat - Initialize translation for StaticColor */
void x_InitSCXlat() {
  int i; 

  if (X_xalloccolor)
    return;

  if (X_visualinfo->depth > SCXLAT_MAXDEPTH) {
    fprintf(stderr, "Reverting to XAllocColor for StaticColor because the "
	    "depth is too high.\n");
    X_xalloccolor |= XALLOCCOLOR_PROGRAM;
    return;
  }
  
  /* Do not compute this from depth.  You can't make that assumption. */
  X_cmapinfoent = X_visualinfo->colormap_size;

  X_cmapinfo = malloc(sizeof(XColor) * X_cmapinfoent);
  if (X_cmapinfo == NULL) {
    fprintf(stderr, "Reverting to XAllocColor for StaticColor because due "
	    "to lack of memory.\n");
    X_xalloccolor |= XALLOCCOLOR_PROGRAM;
    X_cmapinfoent = 0;
    return;
  }

  for (i = 0; i < X_cmapinfoent; i++) {
    X_cmapinfo[i].pixel = i;
  }

  XQueryColors(X_display, X_cmap, X_cmapinfo, X_cmapinfoent);

  for (i = 0; i < X_cmapinfoent; i++) {
    X_cmapinfo[i].red >>= INPAL_SHIFT;
    X_cmapinfo[i].green >>= INPAL_SHIFT;
    X_cmapinfo[i].blue >>= INPAL_SHIFT;
  }

  /* Endiannize the pixel values in X_cmapinfo so we don't need to 
   * endiannize later on
   */
  x_Endiannize(&(X_cmapinfo[0].pixel),
	       ((char *)(&X_cmapinfo[1].pixel) - 
		(char *)(&X_cmapinfo[0].pixel)),
	       X_cmapinfoent);
}
#endif

/* x_AcidwarpToXColors - translates acidwarp palette to an X colormap */ 
void x_AcidwarpToXColors(unsigned char *palette) {  
  /* set the X colormap entries */

  if (X_numcolors == INPAL_SIZE) {
    register unsigned int i;
    register unsigned int c;

    /* Simple 1:1 mapping if we have all the entries we want */
    for (i = 0; i < X_numcolors; i++) {
      c = *palette++;
      X_colors[i].red = (c << INPAL_SHIFT);
      c = *palette++;
      X_colors[i].green = (c << INPAL_SHIFT);
      c = *palette++;
      X_colors[i].blue = (c << INPAL_SHIFT);
    }
  } else {
    /* Where multiple palette entries map to 1 entry we will take an
     * average
     */
    int istart, i, j, n;
    int cend;
    unsigned long int r, g, b;

    /* Special handling for color 0 in Acidwarp */
    if (X_numcolors > 1) {
      istart = 1;
      j = 1;
      X_colors[0].red = *palette++ << INPAL_SHIFT;
      X_colors[0].green = *palette++ << INPAL_SHIFT;
      X_colors[0].blue = *palette++ << INPAL_SHIFT;
    } else {   
      istart = 0;
      j = 0;
    }
    for (i = istart; i < X_numcolors; i++) {
      /* Figure out end of chunk */
      cend = istart + ((i + 1 - istart) * 
		       (INPAL_SIZE - istart) / (X_numcolors - istart));
      n = cend - j;

      /* Compute averages */      
      r = 0; g = 0; b = 0;
      for (; j < cend; j++) {
	r += *palette++;
	g += *palette++;
	b += *palette++;
      }
      X_colors[i].red = (r / n) << INPAL_SHIFT;
      X_colors[i].green = (g / n) << INPAL_SHIFT;
      X_colors[i].blue = (b / n) << INPAL_SHIFT;
    }
  }
}

/* x_UpdateCmapXlat - Updates the colormap on the server */
void x_UpdateCmapXlat(unsigned char *palette) {
  x_AcidwarpToXColors(palette);

#ifdef NO_POINTLESS_STORECOLORS
  if (X_cmapin) {
    /* If our colormap has been installed we then update it */
#endif
    
    /* On a MicroVAX II running NetBSD/vax XStoreColors will update things
     * in color even if the visual is grayscale.  Is that bad?  Should
     * the colors get reset before XStoreColors so things stay gray?
     */
    XStoreColors(X_display, X_cmap, X_colors, X_numcolors); 
#ifdef XF86DGA
    /* Hmm... it seems that unless you do XFDGAInstallColormap after every
     * colormap change the colormap change simply won't take effect if
     * you're running in DGA mode.
     */
    if (X_dga) {
      XF86DGAInstallColormap(X_display, X_screen, X_cmap);
    }
#endif
    XSync(X_display, False);
#ifdef NO_POINTLESS_STORECOLORS
    X_cmapupd = 1;
  } else {
    /* If our colormap hasn't been installed we don't send pointless 
       stuff to the server, but we set a flag so that we update it
       as soon as it is installed. */
    X_cmapupd = 0;
  }
#endif
}

void x_UpdateAllocColorXlat(unsigned char *palette) {
  static int wasallocd = 0;
  int i;

  x_AcidwarpToXColors(palette);
    
  if (wasallocd) 
    XFreeColors(X_display, X_cmap, X_xlattab, X_numcolors, 0);
  
  for (i = 0; i < X_numcolors; i++) {
    XAllocColor(X_display, X_cmap,
		&X_colors[i]);
    X_xlattab[i] = X_colors[i].pixel;
  }
  wasallocd = 1;

  /* Yes, we need to endiannize each time */
  x_Endiannize(&(X_xlattab[0]), sizeof(unsigned long int), INPAL_SIZE);

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

/* x_UpdateColCompXlat - Update translation table with color components */
void x_UpdateColCompXlat(unsigned char *palette)
{
  register unsigned char *ip = palette;
  register unsigned int i;
  register unsigned int t;

  if (X_xalloccolor) {
    x_UpdateAllocColorXlat(palette);
    return;
  }
  
  for (i = 0; i < INPAL_SIZE; i++) { 
    t = X_mapbase;
    t |= X_redmap[*(ip++)];
    t |= X_greenmap[*(ip++)];
    X_xlattab[i] = t | X_bluemap[*(ip++)];
  }
  /* No endiannize needed because X_mapbase, X_redmap, X_greenmap and
   * X_bluemap have all been endiannized
   */

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

/* x_UpdateMonoStaticXlat - Update translation table for StaticGray */
void x_UpdateMonoStaticXlat(unsigned char *palette)
{
  register unsigned char *ip = palette;
  register unsigned int i;
  register unsigned int t;

  if (X_xalloccolor) {
    x_UpdateAllocColorXlat(palette);
    return;
  }
  
  for (i = 0; i < INPAL_SIZE; i++) { 
#if GRAYXLAT_BITS > INPAL_CBITS
    t = *(ip++) << (GRAYXLAT_BITS - INPAL_CBITS);
    t += *(ip++) << (GRAYXLAT_BITS - INPAL_CBITS);
    t += *(ip++) << (GRAYXLAT_BITS - INPAL_CBITS);
#elif GRAYXLAT_BITS < INPAL_CBITS
    t = *(ip++) << (INPAL_CBITS - GRAYXLAT_BITS);
    t += *(ip++) << (INPAL_CBITS - GRAYXLAT_BITS);
    t += *(ip++) << (INPAL_CBITS - GRAYXLAT_BITS);
#else /* GRAYXLAT_BITS == INPAL_CBITS */
    t = *(ip++);
    t += *(ip++);
    t += *(ip++);
#endif
    X_xlattab[i] = X_graymap[t/3];
  }
  /* No endiannizing needed because X_graymap has been endiannized */

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}

#ifdef SCXLAT
void x_UpdateSCXlat(unsigned char *palette) {
  unsigned char *ip = palette;
  int i, j;
  int bestpix;
  int besterr;
  int curerr;
  int r, g, b;
  int dr, dg, db;

  if (X_xalloccolor) {
    x_UpdateAllocColorXlat(palette);
    return;
  }
  
  for (i = 0; i < INPAL_SIZE; i++) { 
    r = *(ip++);
    g = *(ip++);
    b = *(ip++);

    bestpix = 0;
    dr = X_cmapinfo[0].red - r;
    dg = X_cmapinfo[0].green - g;
    db = X_cmapinfo[0].blue - b;
#ifndef SCXLAT_LINEAR
    besterr = dr * dr + dg * dg + db * db;
#else
    besterr = abs(dr) + abs(dg) + abs(db);
#endif
    for (j = 1; j < X_cmapinfoent; j++) {
      dr = X_cmapinfo[j].red - r;
      dg = X_cmapinfo[j].green - g;
      db = X_cmapinfo[j].blue - b;
#ifndef SCXLAT_LINEAR
      curerr = dr * dr + dg * dg + db * db;
#else
      curerr = abs(dr) + abs(dg) + abs(db);
#endif
      if (curerr < besterr) {
	besterr = curerr;
	bestpix = j;
      }
    }
    X_xlattab[i] = X_cmapinfo[bestpix].pixel;
  }

  /* Precalculate shifted translate tables if necessary */
  x_PrecalcXlatShift();
}
#endif

#ifdef XF86DGA
void x_CleanupXF86DGA() {
  XF86DGADirectVideo(X_display, X_screen, 0);
  fprintf(stderr, "Disabled XFree86-DGA direct video.\n");
  
  /* and give back control */
  XUngrabKeyboard(X_display, CurrentTime);
  
  X_dga = 0;

  /* Shutdown the graphics driver */
  X_vhandler->do_cleanup();

  /* Translator functions should be reset just in case */
  X_sxlatfunc = x_SXlatNull;
  X_xlatdest = NULL;
}

#if defined(XF86DGA) && XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
void x_XF86DGASwapBuf() {
  if (X_dblbufbuf == 0) {
    XF86DGASetViewPort(X_display, X_screen, 0, 0);
    X_dblbufbuf = 1;

    X_xlatdest = X_dga_addr + X_height * X_bytes_per_line;
    if (!X_needxlat) 
      buf_graf = X_xlatdest;
  } else {
    XF86DGASetViewPort(X_display, X_screen, 0, X_height);
    X_dblbufbuf = 0;

    X_xlatdest = X_dga_addr;
    if (!X_needxlat) 
      buf_graf = X_xlatdest;
  }
  while(!XF86DGAViewPortChanged(X_display, X_screen, 2)) {}
}
#endif

/* Why can't C handle forward references!?!? */
extern visinfo_t X_visdata[];

void x_InitXF86DGA() {
#define MINDGAMAJOR 0
#define MINDGAMINOR 0
  int MajorVersion, MinorVersion;
  int EventBase, ErrorBase;
  unsigned int width, bank, ram;
  int t;

  /* Verify command line options */
  if (X_dmode & DMODE_ROOT)
    x_Error("XF86DGA can't be used to draw in the root window.\n"); 
  X_dmode |= DMODE_FULLSCREEN;
  if (X_setbg)
    x_Error("XF86DGA doesn't support background pixmap setting.\n");  
  if (X_xputpixel)
    x_Error("XF86DGA doesn't use an XImage and so it can't use XPutPixel.\n");

  if (geteuid()) {
    x_Error("Must be root for XF86DGA.\n");
  }

  if (!XF86DGAQueryVersion(X_display, &MajorVersion, &MinorVersion)) {
    x_Error("Unable to get XFree86-DGA extension version.");
  }

  if (!XF86DGAQueryExtension(X_display, &EventBase, &ErrorBase)) {
    x_Error("Unable to query XFree86-DGA extension.");
  }

  /* Verify command line options */
  /* Fail if the extension version in the server is too old */
  if (MajorVersion < MINDGAMAJOR ||
      (MajorVersion == MINDGAMAJOR && MinorVersion < MINDGAMINOR)) {
    fprintf(stderr,
                "Xserver is running an old XFree86-DGA version"
	    " (%d.%d)\n", MajorVersion, MinorVersion);
    fprintf(stderr, "Minimum required version is %d.%d\n",
	    MINDGAMAJOR, MINDGAMINOR);
    exit(-1);
  }

  if (MajorVersion < 1) {
    int flags;

    XF86DGAQueryDirectVideo(X_display, X_screen, &flags);
    if (!(flags & XF86DGADirectPresent)) {
      x_Error("Xserver driver doesn't support DirectVideo\n");
    }
  }

  /* We always use the default visual */
  t = X_visdeftype;
  X_visdeftype &= VISDEF_ID;
  X_visdeftype |= VISDEF_DEFAULT;
  X_dmode |= DMODE_FULLSCREEN;

  /* We now need to setup stuff.  This is fucked up */
  x_PickVisual();
  X_visdeftype = t;

  /* Note that XF86DGA seems to require a private colormap.  There's no
   * point in not using a private colormap in any case. 
   */
  t = X_cmappref;
  X_cmappref = CMAPPREF_PRIVATE;
  /* Call real init function */
  X_vhandler->do_setup();
  X_cmappref = t;
 
  X_cmapin = 1;
  X_dga = 1;

#ifdef OBSCURED_SLEEP
  /* It's DGA;  It must be visible. */
  X_visstate = VisibilityUnobscured;
  X_mwmapped = 1;
#endif
  X_red_mask = X_visualinfo->red_mask;
  X_green_mask = X_visualinfo->green_mask;
  X_blue_mask = X_visualinfo->blue_mask;

  /* We need this to get X_bits_per_pixel */
  x_GetPixmapFormat(X_visualinfo->depth);

  XF86DGAGetVideo(X_display, X_screen, (char **)(&X_dga_addr), 
                  &width, &bank, &ram);
#if 0
  fprintf(stderr, "XFree86-DGA:  ram:%x, addr %x, banks %d\n",
	  ram, X_dga_addr, bank);
#endif

  X_xlatdest = X_dga_addr;
  X_bytes_per_line = width * X_bits_per_pixel / 8;  
  X_byte_order = ImageByteOrder(X_display);

  /* Get size of window */
  XF86DGAGetViewPortSize(X_display, X_screen,
			 &X_width, &X_height);


#if XF86DGA_DIRW == XF86DGA_DIRW_NEVER || \
(XF86DGA_DBLBUF == XF86DGA_DBLBUF_NEVER && XF86DGA_DIRW == XF86DGA_DIRW_DBLBUF)
  X_needxlat |= NEEDXLAT_BPL;
#else
  /* See if we need to translate due to bytes per line */
  if (X_bytes_per_line != (X_width * X_bits_per_pixel / 8)) {
    X_needxlat |= NEEDXLAT_BPL;
  }
#endif

  /* Do we have enough video RAM for double buffering? */
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
  X_dblbuf = 0;
  X_dblbufbuf = 0;
  if ((2 * X_height * X_bytes_per_line) <= ram * 1024) {
#if XF86DGA_DBLBUF == XF86DGA_DBLBUF_DIRW    
    if (X_needxlat == 0) {
      X_dblbuf = 1;
    } else {
      X_dblbuf = 0;
    }
#else
    X_dblbuf = 1;
#endif
  }
  
  if ((2 * X_height * X_bytes_per_line) > bank) {
#if XF86DGA_DBLBUF == XF86DGA_DBLBUF_DIRW    
    X_dblbuf = 0;
#else
    if (X_needxlat == 0) {
      /* Don't spoil direct writes by double-buffering copies */
      X_dblbuf = 0;
    }
#endif
  }
#endif

  /* Can we still do direct writes if we don't have double buffering */
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER && \
    XF86DGA_DIRW == XF86DGA_DIRW_DBLBUF
  if (!X_dblbuf) {
    X_needxlat |= NEEDXLAT_BPL;
  }
#endif

  if ((
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
       (1 + X_dblbuf) *
#endif
       X_height * X_bytes_per_line) > bank) {
    x_Error("Banked XFree86-DGA modes not supported.  Sorry.\n");
  }

  if (!X_needxlat) {
    buf_graf = X_xlatdest;
  } else {
    /* We do need to translate */
    buf_graf = calloc(X_width * X_height * X_bits_per_pixel / 8, 1);
    if (buf_graf == NULL) {
      x_Error("Failed to allocate Acidwarp graphics buffer.\n");
    }

    x_SelectXlatFunc();
  }

  XGrabKeyboard(X_display, X_rootwin, True, GrabModeAsync,
		GrabModeAsync,  CurrentTime);

  fprintf(stderr, "XFree86-DGA:  %s, %s, UnBanked, Handler = %s\n",
	  X_needxlat ? "Translated" : "DirectWrite",
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
	  X_dblbuf ? "DoubleBuffer" :
#endif
	  "SingleBuffer",
	  X_vhandler->name);

  XSync(X_display, False);   /* Dump errors before the risk */
  fprintf(stderr, "Starting XFree86-DGA DirectGraphics.\n");
  XF86DGADirectVideo(X_display, X_screen,
		     XF86DGADirectGraphics |
		     XF86DGADirectKeyb); 

#ifndef __EMX__
  /* Give up root privs */
  setuid(getuid());
#endif

  if (X_colormapped)
    XF86DGAInstallColormap(X_display, X_screen, X_cmap);

  /* If double-buffered, display it.  Else just set viewport */
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
  if (X_dblbuf) {
    x_XF86DGASwapBuf();
    memset(X_dga_addr, 0, 2 * X_bytes_per_line * X_height);
  } else {
#endif
    XF86DGASetViewPort(X_display, X_screen, 0, 0);
    memset(X_dga_addr, 0, X_bytes_per_line * X_height);
#if XF86DGA_DBLBUF != XF86DGA_DBLBUF_NEVER
  }
#endif
}
#endif

/* Emulation of non-colormapped mode using a colormap */
void x_InitCmapFixEm() {
  /* We *are* using a colormap */
  x_InitColormapped();
  
  /* Colormap is now fixed, so we need translation */
  X_needxlat |= NEEDXLAT_FIXCMAP;
}

/* Initialize colormap and translation table for StaticGray emulation in 
   GrayScale */
void x_InitGSSGEXlat() {
  unsigned int i, j, t, cend;

  /* Fill out palette entries */
  j = 0;
  for (i = 0; i < X_numcolors; i++) {
    /* Figure out end of chunk */
    cend = ((i + 1) * GRAYXLAT_SIZE / X_numcolors);

    /* Figure out palette entry (stretched) */
    t = i * 0xffff / (X_numcolors - 1);

    X_colors[i].red = t;
    X_colors[i].green = t;
    X_colors[i].blue = t;
#if 0
    fprintf(stderr, "[%i] = %i : %i-%i = %i\n", 
	    i, X_colors[i].pixel, j, cend-1, t);
#endif
#ifndef CLR_RELLOC_RDONLY
    /* Fill out chunk in gray translation table */
    for (; j < cend; j++) {
      X_graymap[j] = X_colors[i].pixel;
    }
#endif
    j = cend; 
  }
  
  /* Store our now static colormap */
#ifdef CLR_REALLOC_RDONLY
  if (!X_privatecmap)
    x_ClrReallocRdonly();
  else
#endif
    XStoreColors(X_display, X_cmap, X_colors, X_numcolors); 

#ifdef CLR_REALLOC_RDONLY
  /* Fill out the graymap */
  j = 0;
  for (i = 0; i < X_numcolors; i++) {
    /* Figure out end of chunk */
    cend = ((i + 1) * GRAYXLAT_SIZE / X_numcolors);

    /* Fill out chunk in gray translation table */
    for (; j < cend; j++) {
      X_graymap[j] = X_colors[i].pixel;
    }
    j = cend;  /* pointless */ 
  }
#endif
}

#ifdef SCXLAT 
void x_InitPCSCEXlat() {
  int n, t;
  int rplanes, gplanes, bplanes;
  int numred, numgreen, numblue;
  int r, g, b;

  /* Number of planes to put into all 3 */
  n = 0;
  t = X_numcolors;
  while (1) {
    t >>= 3;
    if (t == 0) break;
    n++;
  }

  gplanes = n;
  rplanes = n;
  bplanes = n;

#if 0
  /* Take one more plane for green if possible */
  t = X_numcolors >> (n * 3);
  if (t & ~1) {
    gplanes++;
    /* Take one more plane for red if possible */
    if (t & ~3) {
      rplanes++;
    }
  }
#endif

  /* Calculate number of remaining colors */
  t = 1 << (rplanes + gplanes + bplanes);
  n = X_numcolors - t;
  
  /* Calculate number of distinct values */
  numred = 1 << rplanes;
  numblue = 1 << bplanes;
  numgreen = 1 << gplanes;

  while (1) {
    /* Try to get more colors */
    t = numred * numgreen;
    if (numblue < numred && n >= t) {
      numblue++;
      n -= t;
      continue;
    }
    t = numblue * numgreen;
    if (numred < numgreen && n >= t) {
      numred++;
      n -= t;      
      continue;
    }
    t = numblue * numred;
    if (n >= t) {
      numgreen++;
      n -= t;
      continue;
    }
    break;
  }

  /* Check if we have enough colours */
  if (numred < 2 || numgreen < 2 || numblue < 2) {
    x_Error("Insufficient colormap entries for StaticColor emulation.\n");
  }

  /* Fill out the cube */
  t = 0;
  for (r = 0; r < numred; r++) {
    rplanes = r * 0xffff / (numred - 1);
    for (g = 0; g < numgreen; g++) {
      gplanes = g * 0xffff / (numgreen - 1);
      for (b = 0; b < numblue; b++) {
	X_colors[t].red = rplanes;
	X_colors[t].green = gplanes;
	X_colors[t].blue = b * 0xffff / (numblue - 1);
	t++;
      }
    }
  }

  fprintf(stderr, "StaticColor emulation in PseudoColor, RxGxB = %ix%ix%i\n",
	  numred, numgreen, numblue);

  if (t > X_numcolors) {
    /* Just to aid testing */
    fprintf(stderr, "t = %i, X_numcolors = %i\n", t, X_numcolors);
    x_Error("x_InitPCSCEXlat algorithm error\n");
  }

  if (!X_privatecmap) {
    /* Free any remaining colors */
    for (n = t; n < X_numcolors; n++) {
      XFreeColors(X_display, X_cmap, &X_colors[n].pixel, 1, 0);
      XSync(X_display, False);
    }

    X_numcolors = t;
#ifdef CLR_REALLOC_RDONLY
    x_ClrReallocRdonly();
#else
    XStoreColors(X_display, X_cmap, &X_colors[0], t);
#endif
  } else {
    XStoreColors(X_display, X_cmap, &X_colors[0], t);
  }
  
  X_numcolors = t;

  /* Fill out the translation table */
  X_cmapinfoent = X_numcolors;
  X_cmapinfo = X_colors;
  for (t = 0; t < X_cmapinfoent; t++) {
    X_cmapinfo[t].red >>= INPAL_SHIFT;
    X_cmapinfo[t].green >>= INPAL_SHIFT;
    X_cmapinfo[t].blue >>= INPAL_SHIFT;
#if 0
    fprintf(stderr, "[%i] = %i,%i,%i\n", t,
	    X_colors[t].red, X_colors[t].green, X_colors[t].blue);
#endif
  }

  x_Endiannize(&(X_cmapinfo[0].pixel),
	       ((char *)(&X_cmapinfo[1].pixel) - 
		(char *)(&X_cmapinfo[0].pixel)),
	       X_cmapinfoent);
}
#endif

/* These cleanup functions should eventually be good enough so that no 
 * matter when they're called (eg. after an error) they will clean up
 * whatever needs to be cleaned up.
 */

void x_CleanupColormapped() {
  if (X_privatecmap) {
    if (X_cmap != None) {
      XFreeColormap(X_display, X_cmap);
      X_cmap = None;
    }
  } else {
    x_FreeColors();
  }
}

void x_CleanupFixCMap() {
  if (X_privatecmap && X_cmap != None) {
    XFreeColormap(X_display, X_cmap);
  }
}

#ifdef SCXLAT
void x_CleanupSCXlat() {
  if (X_cmapinfo != NULL && X_cmapinfo != X_colors) {
    free(X_cmapinfo);
    X_cmapinfoent = 0;
    X_cmapinfo = NULL;
  }

  x_CleanupFixCMap();
}

void x_CleanupPCSCE() {
  /* Undo endiannize so the pixels will be right */
  x_Endiannize(&(X_cmapinfo[0].pixel),
	       ((char *)(&X_cmapinfo[1].pixel) - 
		(char *)(&X_cmapinfo[0].pixel)),
	       X_cmapinfoent);
  
  x_CleanupColormapped();
  
  X_cmapinfoent = 0;
  X_cmapinfo = NULL;
}
#endif

void x_CleanupDCTCE() {
  if (X_privatecmap) {
    if (X_cmap != None) {
      XFreeColormap(X_display, X_cmap);
    }
  } else {
    if (X_red_mask || X_green_mask || X_blue_mask) {
      XFreeColors(X_display, X_cmap, &X_mapbase, 1, 
		  X_red_mask | X_green_mask | X_blue_mask);
      X_red_mask = 0;
      X_green_mask = 0;
      X_blue_mask = 0;
    }
  }
}


visinfo_t X_visdata[] = {
  { PseudoColor, "PseudoColor", "Great performance, colormapped",
    VISINFO_COLORMAPPED, 8, 
    x_InitColormapped, x_InitCmapXlat, x_UpdateCmapXlat,
    x_CleanupColormapped
  },
  { TrueColor, "TrueColor", "Bad performance",
    0, 24, 
    x_InitFixCMap, x_InitTrueColorXlat, x_UpdateColCompXlat,
    x_CleanupFixCMap 
  },
  { DirectColor, "DirectColorTCE", "TrueColor emulation on DirectColor",
    VISINFO_COLORMAPPED, 8, 
    x_InitDCTCE, x_NullFunction, x_UpdateColCompXlat,
    x_CleanupDCTCE
  },
  { StaticColor, "StaticColor", "Bad performance, limited colors", 
    0, 16, 
#ifdef SCXLAT
    x_InitFixCMap, x_InitSCXlat, x_UpdateSCXlat,
    x_CleanupSCXlat
#else
    x_InitFixCMap, x_NullFunction, x_UpdateAllocColorXlat,
    x_CleanupFixCMap
#endif
  },
  { DirectColor, "DirectColorPCE", "PseudoColor emulation on DirectColor",
    VISINFO_COLORMAPPED, 24, 
    x_InitColormapped, x_InitCmapXlat, x_UpdateCmapXlat,
    x_CleanupColormapped
  },
#ifdef SCXLAT
  { PseudoColor, "PseudoColorSCE", "StaticColor emulation on PseudoColor",
    VISINFO_COLORMAPPED, 4,
    x_InitCmapFixEm, x_InitPCSCEXlat, x_UpdateSCXlat,
    x_CleanupPCSCE
  },
#endif
  { GrayScale, "GrayScale", "Great performace, colormapped, monochrome", 
    VISINFO_COLORMAPPED | VISINFO_MONO, 8, 
    x_InitColormapped, x_InitCmapXlat, x_UpdateCmapXlat,
    x_CleanupColormapped
  },
  { StaticGray, "StaticGray", "Bad performance, monochrome",
    VISINFO_MONO, 8, 
    x_InitFixCMap, x_InitStaticGrayXlat, x_UpdateMonoStaticXlat,
    x_CleanupFixCMap
  },
  { GrayScale, "GrayScaleSGE", "StaticGray emulation in GrayScale", 
    VISINFO_COLORMAPPED | VISINFO_MONO, 4, 
    x_InitCmapFixEm, x_InitGSSGEXlat, x_UpdateMonoStaticXlat,
    x_CleanupColormapped
  }
};
#define NUM_VIS (sizeof(X_visdata)/sizeof(visinfo_t))

/* x_PickVisual - Selects a visual */
void x_PickVisual() {
  XVisualInfo *visualinfo, vist, *usevis = NULL;
  XVisualInfo *candidates[NUM_VIS];
  int i = 0, j, n = 0, mask;
  int preselhand = 0;

  if (!(X_dmode & DMODE_FULLSCREEN) &&
      (X_visdeftype & VISDEF_HANDLER)) {
    i = X_visdef_handler;
    preselhand = 1;
  } else if ((X_dmode & DMODE_FULLSCREEN) &&
	     (X_visdeftype & VISDEF_FSHAND)) {
    i = X_visdef_fshand;
    preselhand = 1;
  }

  if (preselhand) {
    if (i < 0 || i >= NUM_VIS) {
      x_Error("Invalid pre-selected visual handler.  Use -v ? to get a "
	      "list.\n");
    } else if (X_visdata[X_visdef_handler].flags & VISINFO_NOTXWIN) {
      /* This is *not* a standard X11 handler */
      X_vhandler = &X_visdata[X_visdef_handler];
      X_colormapped = X_vhandler->flags & VISINFO_COLORMAPPED;
      return;
    }
  }

  /* Get info on all visual classes we care about for this screen */
  mask = VisualScreenMask;
  vist.screen = X_screen;
  if (X_visdeftype & VISDEF_ID) {
    mask |= VisualIDMask,
    vist.visualid = X_visdef_id;
  } else   /* If the default visual was selected get it's id */
  if (X_visdeftype & VISDEF_DEFAULT) {
    mask |= VisualIDMask,
    vist.visualid = XVisualIDFromVisual(DefaultVisual(X_display, X_screen));
  }

  visualinfo = XGetVisualInfo(X_display, mask, &vist, &n);
  if (X_visdeftype & VISDEF_ID) {
    if (n <= 0) {
      x_Error("Selected visual class not found.  Use xdpyinfo to get info on"
	      " your X server.\n");
    } else if (n != 1) {
      fprintf(stderr, "warning: Searching for specific visual returned "
	      "multiple results.\n");
    }
  } else if (n <= 0) {
    x_Error("No visuals found.  Use xdpyinfo to get info on your "
	      "X server.\n");
  }

  /* Scan through all the visuals that we support, or i can be set to 
   * the particular handler we want.
   */
  for (; i < NUM_VIS; i++) {
    candidates[i] = NULL;
    if (!(X_visdata[i].flags & VISINFO_ALIEN)) {       
      /* For each one we support scan through all available visuals to find
	 all instances of that visual */
      for (j = 0; j < n; j++) {
	if (visualinfo[j].class == X_visdata[i].class) {
	  /* Okay, we can handle this visual */
	  /* If we haven't found this one before... */
	  if (candidates[i] == NULL || 
	      /* or if the best depth we have found so far is suboptimal... */
	      (candidates[i]->depth < X_visdata[i].wantdepth &&
	       visualinfo[j].depth > X_visdata[i].wantdepth) ||
	      /* or the best so far is too high but both are high enough */
	      (candidates[i]->depth > X_visdata[i].wantdepth &&
	       visualinfo[j].depth > X_visdata[i].wantdepth &&
	       visualinfo[j].depth < candidates[i]->depth) ) {
	    /* Then this visual is the new candidate */ 
	    candidates[i] = &(visualinfo[j]);
	  }
	}
      }
      
      /* If we got a depth or if this was the preselected visual then
	 we use this visual */
      if (candidates[i] != NULL && (preselhand ||
	    candidates[i]->depth >= X_visdata[i].wantdepth)) {
	usevis = candidates[i];
	break;
      }
      
      if (preselhand) {
	/* A visual has been preselected */
	if (X_visdeftype & VISDEF_ID) {
	  x_Error("The selected visual handler is not compatible with the "
		  "selected \nvisual class.\n");
	} else if (X_visdeftype & VISDEF_DEFAULT) {
	  x_Error("The selected visual handler wasn't compatible with the"
		  " Default visual.\n");
	} else {
	  x_Error("You have preselected a visual handler that is not "
		  "supported by your X server\nUse xdpyinfo to get info "
		  "on your server.\n");
	}
      }
    }
  }
  
  /* The scan hasn't selected a visual */      
  if (usevis == NULL) {
    /* We didn't find the depth that we want.  Now we'll just pick
       the first visual that we can use regardless of depth.
       Perhaps this should pick one with the highest depth or something. */
    for (i = 0; i < NUM_VIS; i++) {
      if (candidates[i] != NULL) {
	usevis = candidates[i];
	break;
      }
    }
    
    /* If we still haven't managed to find a visual then there just isn't
       any supported visual. */
    if (usevis == NULL) {
      x_Error("No supported visuals found.  Use the -v ? switch to list"
	      " visuals supported by\nthis program and use xdpyinfo to "
	      "get info on your X server.\n");
    }
  }
  
  /* Now save info about this visual. */
  X_visualinfo = XGetVisualInfo(X_display, VisualIDMask, usevis, &n);
  if (X_visualinfo == NULL || n != 1) {
    x_Error("Error getting info for selected visual.");
  }
  XFree(visualinfo);
  X_visual = X_visualinfo->visual; 

  /* Leave pointer to data about our capabilities on this visual */
  X_vhandler = &X_visdata[i];
  
  /* Set colormapped flag appropriately */
  X_colormapped = X_vhandler->flags & VISINFO_COLORMAPPED;
}

/* List visuals */
void x_ListVis() {
  int i;
  
  fprintf(stderr, "Visual classes supported by Acidwarp:\n\n"
	  "Class        WantBPP  Comment\n");
  for (i = 0; i < NUM_VIS; i++) {
    fprintf(stderr, "%-17s %2u  %s\n", X_visdata[i].name,
	    X_visdata[i].wantdepth, X_visdata[i].comment);
  }
}

/* Return visual number from visual name */
int x_SetWantedVis(char *s, unsigned int mask) {
  int t;
 
  if (!strcmp("?", s)) {
    x_ListVis();
    exit(0);
  }

  if (X_visdeftype & mask) {
    x_Error("You can only select a visual handler for a mode once.\n");
  }

  if (*s != 0) {
    for (t = 0; t < NUM_VIS; t++) {
      if (!strcmp(X_visdata[t].name, s)) {
	X_visdeftype |= mask;
	return t;
      }
    }
  }
  
  x_Error("Bad visual handler selection.  Use -v ? to list handlers.\n");

  /* Stop complaining about not returning a value */
  return 666;
}     

