/* xxlat.c - Translation routines for LibXPCE, the X PseudoColor emulation 
 *           library.  This file deals with actually translating the 
 *           8 bit PseudoColor data into the format needed for displaying
 *           it on the X server
 *
 * Copyright (C) 2000 Boris Gjenero
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1, as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You may contact Boris Gjenero at bgjenero@sympatico.ca
 */

#include <stdio.h>
#include "xconfig.h"
#include "x.h"

/********************** GLOBAL VARIABLES *************************************/

int X_needxlat;
unsigned long int X_xlattab1[INPAL_SIZE];
unsigned long int X_xlattab2[INPAL_SIZE];
unsigned long int X_xlattab3[INPAL_SIZE];
unsigned long int X_xlattab4[INPAL_SIZE];

unsigned char *X_xlatdest;

void (*X_sxlatfunc)(unsigned char *,  /* Destination */ 
		    unsigned int, unsigned int,  /* x, y */
		    unsigned int, unsigned int   /* width, height */
		    ) = x_SXlatNull;

/*********************** TRANSLATOR FUNCTIONS ********************************/

/* NOTE:  The translator used *only* depends on the bits per pixel of the
 *        XImage, and *nothing* else.  Anything may be accomplished by
 *        appropriately using the translator and translate table.
 */

/* Null translator function that just reports an error */
void x_SXlatNull(unsigned char *d, 
		 unsigned int x, unsigned int y, 
		 unsigned int w, unsigned int h) {
  x_Error("Generallized translator function not set up.\n");
}

/* Translate where there's just a different number of bytes per line */
void x_SXlatBPL(unsigned char *d, unsigned int x, unsigned int y, 
		unsigned int w, unsigned int h) {
  int i;
  unsigned char *ip = buf_graf + x + (y * X_width);
  unsigned char *op = d;
  
  for (i = 0; i < h; i++) {
    /* Here the assumption is that memcpy can do pure copying better than
       a C loop.  This assumption may be wrong. */
    memcpy(op, ip, w);
    ip += X_width;
    op += X_bytes_per_line;
  }
}

/* Generallized 8BPP translator function.  Note that all the data is actually
 * read and written unsigned long ints and not chars, because that's faster.
 * Performance depends on alignment.  If the starting position of each
 * row in the source and destination isn't mod 4 aligned, then you have
 * a problem.
 *
 * Endianness - We get a long int and we put a long int, so it cancels
 *   out.  Make sure that X_xlattab1 deals with 0xff, 2 with 0xff00,
 *   3 with 0xff0000 and 4 with 0xff000000.  No server byte ordering
 *   issues exist.
 */
void x_SXlat8BPP(unsigned char *d, unsigned int x, unsigned int y, 
		unsigned int w, unsigned int h) {
  unsigned int i;
  unsigned char *nextip, *nextop;
  register unsigned long int t;
  register unsigned char *ipend;
  register unsigned char *ip = buf_graf + x + (y * X_width);
  register unsigned char *op = d;
  
  for (i = 0; i < h; i++) {
    /* Do divisible by 4 part */
    ipend = ip + (w & ~3);
    nextop = op + X_bytes_per_line;
    nextip = ip + X_width;
    while (ip < ipend) {
      t = *(unsigned long int *)ip;
      *(unsigned long int*)op = 
	   X_xlattab1[t & 0xff]
	| (X_xlattab2[(t >> 8) & 0xff])
	| (X_xlattab3[(t >> 16) & 0xff])    
	| (X_xlattab4[t >> 24]);

      ip += 4;
      op += 4;
    }

    /* Handle remainder */
    ipend = ip + (w & 3);    
    while (ip < ipend) {
      *((unsigned char *)(op)++) = X_xlattab[*(ip++)];
    }
    
    /* Reset for next line */
    op = nextop;
    ip = nextip;
  }
}



/* Generallized 16BPP translate function.  Note that all the data is actually
 * read and written unsigned long ints and not chars, because that's faster.
 * Performance depends on alignment.  If the starting position of each
 * row in the source and destination isn't mod 4 aligned, then you have
 * a problem.
 *
 * Endianness - We are getting 4 bytes at a time and so host endianness
 *   affects us.  This may be dealt with using compile time defines.  
 *   Pixels are 2 bytes, and so byte_order affects us.  This may be dealt
 *   with by swapping the translation table before translating.  Host
 *   endianness also gets us when we're storing, but compile-time defines
 *   deal with that too.
 */

void x_SXlat16BPP(unsigned char *d, unsigned int x, unsigned int y, 
		unsigned int w, unsigned int h) {
  unsigned int i;
  unsigned char *nextop, *nextip;
  register unsigned long int t;
  register unsigned char *ipend;
  register unsigned char *ip = buf_graf + x + (y * X_width);
  register unsigned char *op = d;
  
  for (i = 0; i < h; i++) {
    /* Setup for next line */
    nextop = op + X_bytes_per_line;
    nextip = ip + X_width;

    /* Do divisible by 4 part */
    ipend = ip + (w & ~3);
    nextip = ip + X_width;
    while (ip < ipend) {
      /* Get 4 pixels */
      t = *(unsigned long int *)ip;

      /* Deal with host byte order */
#ifdef XXLAT_BIGENDIAN
      /* - Big Endian - most significant byte first */
      *(unsigned long int*)op = 
	X_xlattab2[t >> 24] |
	X_xlattab1[(t >> 16) & 0xff];
      *(unsigned long int*)(op + 4) = 
	X_xlattab2[(t >> 8) & 0xff] |
	X_xlattab1[t & 0xff];
#else
      /* - Little Endian - least significant byte first */
      *(unsigned long int*)op = 
	X_xlattab1[t & 0xff] |
	X_xlattab2[(t >> 8) & 0xff];
      *(unsigned long int*)(op + 4) = 
	X_xlattab1[(t >> 16) & 0xff] |
	X_xlattab2[t >> 24];
#endif
      ip += 4;
      op += 8;
    }

    /* Handle remainder */
    ipend = ip + (w & 3);    
    while (ip < ipend) {
      *(unsigned short int *)(op) = X_xlattab1[*(ip++)];
      op += 2;
    }

    /* Reset for next line */
    op = nextop;
    ip = nextip;
  }
}

/* Generallized 32BPP translate function.  
 * Performance depends on alignment.  If the starting position of each
 * row in the source and destination isn't mod 4 aligned, then you have
 * a problem.
 *
 * Endianness - We are getting 4 bytes at a time and so host endianness
 *   affects us.  This may be dealt with using compile time defines.  
 *   Pixels are 4 bytes, and so byte_order affects us.  This may be dealt
 *   with by swapping the translation table before translating.
 */
void x_SXlat32BPP(unsigned char *d, unsigned int x, unsigned int y, 
		unsigned int w, unsigned int h) {
  unsigned int i;
  unsigned char *nextop, *nextip;
  register unsigned long int t;
  register unsigned char *ipend;
  register unsigned char *ip = buf_graf + x + (y * X_width);
  register unsigned char *op = d;

  for (i = 0; i < h; i++) {
    /* Do divisible by 4 part */
    ipend = ip + (w & ~3);
    nextop = op + X_bytes_per_line;
    nextip = ip + X_width;
    while (ip < ipend) {
      /* Get 4 pixels */
      t = *(unsigned long int *)ip;
      /* Deal with host byte order */
#ifdef XXLAT_BIGENDIAN
      /* - Big Endian - most significant byte first */
      *(unsigned long int*)op = X_xlattab[t >> 24];
      *(unsigned long int*)(op + 4) = X_xlattab[(t >> 16) & 0xff];
      *(unsigned long int*)(op + 8) = X_xlattab[(t >> 8) & 0xff];
      *(unsigned long int*)(op + 12) = X_xlattab[t & 0xff];
#else
      /* - Little Endian - least significant byte first */
      *(unsigned long int*)op = X_xlattab[t & 0xff];
      *(unsigned long int*)(op + 4) = X_xlattab[(t >> 8) & 0xff];
      *(unsigned long int*)(op + 8) = X_xlattab[(t >> 16) & 0xff];
      *(unsigned long int*)(op + 12) = X_xlattab[t >> 24];
#endif
      ip += 4;
      op += 16;
    }

    /* Handle remainder */
    ipend = ip + (w & 3);    
    while (ip < ipend) {
      *(unsigned long int *)(op) = X_xlattab[*(ip++)];
      op += 4;
    }
    
    op = nextop;
    ip = nextip;
  }
}



/* Translator function using XPutPixel.  Note that in this case the
 * destination is ignored.
 */
void x_SXlatPP(unsigned char *d, unsigned int x, unsigned int y, 
	       unsigned int w, unsigned int h) {
  int i, j;
  unsigned char *nextip, *ip = buf_graf + x + (y * X_width);

  for (i = y; i < y + h; i++) {
    nextip = ip + X_width;
    for (j = x; j < x + w; j++) {
      XPutPixel(X_image, j, i, X_xlattab[*ip]); 
      ip++;
    }
    ip = nextip;
  }
}

/* Translator function using XPutPixel where only bytes per row differs.
 * Note that in this case the destination is ignored.
 */
void x_SXlatPP_BPL(unsigned char *d, unsigned int x, unsigned int y, 
		   unsigned int w, unsigned int h) {
  int i, j;
  unsigned char *nextip, *ip = buf_graf + x + (y * X_width);

  for (i = y; i < y + h; i++) {
    nextip = ip + X_width;
    for (j = x; j < x + w; j++) {
      XPutPixel(X_image, j, i, *ip); 
      ip++;
    }
    ip = nextip;
  }
}


/************************* GENERAL FUNCTIONS *********************************/

/* Generallized endianness crap for bits_per_pixel % 8 == 0
 *
 * You have:  bit_order of LSBFirst or MSBFirst, which determines the
 *              order of bytes in a multi-byte pixel and nothing else.
 *
 *            host endianness which defines the order of stuff we load
 *              and store.
 */

void x_Endiannize(unsigned long int *start,
		  unsigned int spacing,
		  unsigned int n) {
  register unsigned long int t;
  register unsigned long int *p;
  register unsigned long int *endp;
  register unsigned int s;
  
  if(
#ifdef XXLAT_ENDIAN
     X_byte_order == LSBFirst
#else
     X_byte_order == MSBFirst
#endif  
     ) {
    
    switch (X_bits_per_pixel) {
    case 32:
      p = start;
      endp = (unsigned long int *)((unsigned char *)start + (spacing * n));
      s = spacing;

      while (p < endp) {
	t = *p;
	*p = 
	  (t << 24) | 
	  ((t << 8) & 0x00ff0000) |
	  ((t >> 8) & 0x0000ff00) |
	  (t >> 24);
	p = (unsigned long int *)((unsigned char *)p + spacing);
      }
      break;
    case 16:
      p = start;
      endp = (unsigned long int *)((unsigned char *)start + (spacing * n));
      s = spacing;

      while (p < endp) {
	t = *p;
	*p = 
	  ((t << 8) & 0xff00) |
	  (t >> 8);
	  p = (unsigned long int *)((unsigned char *)p + spacing);
      }
      
      break;
    }
  }
}

/* x_PrecalcXlatShift - Calculate shifted translation tables.  I have found 
 *                      that this is faster than acutally doing the shift
 *                      in the translation routine
 */
void x_PrecalcXlatShift() {
  register unsigned int i, t;

  switch (X_bits_per_pixel) {
  case 8:
    for (i = 0; i < INPAL_SIZE; i++) {
      t = X_xlattab[i];
#ifndef X_xlattab
      X_xlattab1[i] = t;
#endif
      X_xlattab2[i] = t << 8;
      X_xlattab3[i] = t << 16;
      X_xlattab4[i] = t << 24;
    }
    break;
  case 16:
    for (i = 0; i < INPAL_SIZE; i++) {
#ifndef X_xlattab
      t = X_xlattab[i];
      X_xlattab1[i] = X_xlattab[i];
      X_xlattab2[i] = t;
#else
      X_xlattab2[i] = X_xlattab[i] << 16;
#endif
    }
    break;
  }
}


/* This is a simple "translate whole buffer" function */
inline void x_XlatBuffer() {
  X_sxlatfunc(X_xlatdest, 0, 0, X_width, X_height);
}

/* Function that sets up translator */
void x_SelectXlatFunc() {
  /* If PutPixel translates are being used, just select the appropriate one
   * and exit.
   */
  if (X_xputpixel) {
    if (X_image == NULL) {
      x_Error("No XImage exists, so XPutPixel can't be used.\n");
    }
    if (X_needxlat == NEEDXLAT_BPL) {
      X_sxlatfunc = x_SXlatPP_BPL;
    } else {
      X_sxlatfunc = x_SXlatPP;
    }
    return;
  }

  /* Set complete translator function */
  X_sxlatfunc = NULL;
  if (X_needxlat == NEEDXLAT_BPL) {
    /* Only bytes per line differ. */
    X_sxlatfunc = x_SXlatBPL;
  } else {
    switch (X_bits_per_pixel) {
    case 32:
      X_sxlatfunc = x_SXlat32BPP;
      break;
    case 16:
      X_sxlatfunc = x_SXlat16BPP;
      break;
    case 8:
      X_sxlatfunc = x_SXlat8BPP;
      break;
    default:
      if (X_image != NULL) {
	fprintf(stderr, "No optimized translator function for %i bits per "
		"pixel.  Reverting\nto XPutPixel.\n", X_bits_per_pixel);
	X_sxlatfunc = x_SXlatPP;
      }
    }
  }
  
  if (X_sxlatfunc == NULL) {
    fprintf(stderr, "No complete translator function for %i BPP\n", 
	    X_bits_per_pixel);
    exit(-1);
  }
}	
