/* xmms-vigor : A Vigor visualization plugin for XMMS
 * Copyright (C) 2003 Joel Ray Holveck
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <xmms/util.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <stdio.h>
#include <xmms/plugin.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <SDL/SDL_syswm.h>
#include <X11/Xlib.h>

#include "speech-bubble.h"

/* Forward declarations */
static void vigor_init(void);
static void vigor_cleanup(void);
static void vigor_playback_start(void);
static void vigor_playback_stop(void);
static void vigor_render_freq(gint16 freq_data[2][256]);

/* Callback functions */
static VisPlugin vigor_vtable = {
  0, /* Handle, filled in by xmms */
  0, /* Filename, filled in by xmms */

  0,                    /* Session ID */
  "Vigor Plugin",       /* description */

  0, /* # of PCM channels for render_pcm() */
  2, /* # of freq channels wanted for render_freq() */

  vigor_init,			/* Called when plugin is enabled */
  vigor_cleanup,		/* Called when plugin is disabled */
  0,				/* Show the about box */
  0,				/* Show the configure box */
  0,		     /* Called to disable plugin, filled in by xmms */
  vigor_playback_start,		/* Called when playback starts */
  vigor_playback_stop,		/* Called when playback stops */
  0,			/* Render the PCM data, must return quickly */
  vigor_render_freq    /* Render the freq data, must return quickly */
};

/* XMMS entry point */
VisPlugin
*get_vplugin_info(void)
{
  return &vigor_vtable;
}

/* Global vars for our worker thread */
static SDL_Thread *worker_thread;
static enum { GO, STOP } thread_control;
static SDL_mutex *bv_mutex, *bubble_mutex;
static gint16 bass_value;
static SDL_Surface *screen;

/* Images */
static SDL_Surface *img_vigor_empty, *img_vigor_rest,
  *img_vigor_ent_0001, *img_vigor_ent_0002, *img_vigor_ent_0003,
  *img_vigor_ent_0004, *img_vigor_wav_0001,
  *img_vigor_wav_0002;

#define SAMPLE_AVG_PERIOD 16
#define SAMPLE_AVG_PERIOD_LOG2 4
static gint16 bass_samples[SAMPLE_AVG_PERIOD];
static unsigned int ringpos;

/* usecs between samples */
#define SAMPLE_TIME 75000

static void
worker_func(void)
{
  gint16 bv;
  SDL_Surface *s;
  SDL_Surface *ls = NULL;

  for (ringpos = 0; ringpos < SAMPLE_AVG_PERIOD; ringpos++)
    bass_samples[ringpos] = 0;
  ringpos = 0;

  while (thread_control == GO) {
    unsigned int i;
    unsigned int mean, hv, lv;	/* Mean, high range, low range */

    SDL_mutexP(bv_mutex);
    bv = bass_value;
    SDL_mutexV(bv_mutex);

    bass_samples[ringpos++] = bv;
    if (ringpos >= SAMPLE_AVG_PERIOD) 
      ringpos = 0;

    mean = 0;
    for (i = 0; i < SAMPLE_AVG_PERIOD; i++)
      mean += bass_samples[i] >> SAMPLE_AVG_PERIOD_LOG2;
    hv = mean * 1.2;
    lv = mean * 0.8;

    if (bv > hv)
      s = img_vigor_wav_0002;
    else if (bv > lv)
      s = img_vigor_wav_0001;
    else
      s = img_vigor_rest;
    if (s != ls) {
      ls = s;
      SDL_BlitSurface(s, NULL, screen, NULL);
      SDL_UpdateRect(screen, 0,0,0,0);
    }

    xmms_usleep(SAMPLE_TIME);
  }
#if 0
  printf("Worker thread: Exiting\n");
#endif
}

#define xstr(s) str(s)
#define str(s) #s
static int
initialize_image(SDL_Surface **image, const char* fname) 
{
  char *path;
  int pathlen, ppr;

  pathlen = strlen(xstr(PKGDATADIR) "/.bmp") + strlen(fname) + 1;
  path = malloc(pathlen);
  if (path == NULL)
    return -1;
  /* Don't put PKGDATADIR directly in the string, or a % in there will lose. */
  ppr = snprintf(path, pathlen, "%s/%s.bmp", xstr(PKGDATADIR), fname);
  if (ppr != (pathlen-1)) {
    printf ("canthappen at %s:%d\n", __FILE__, __LINE__);
    free(path);
    return -1;
  }

  *image = SDL_LoadBMP(path);
  free(path);

  return (*image == NULL) ? -1 : 0;
}

static void
vigor_init(void) 
{
  if (SDL_Init (SDL_INIT_VIDEO) < 0) {
    printf ("Failed to initialize SDL\n");
    vigor_vtable.disable_plugin (&vigor_vtable);
    return;
  }

  /* Initialize our images */
  if (initialize_image(&img_vigor_empty, "vigor_empty") ||
      initialize_image(&img_vigor_rest, "vigor_rest") ||
      initialize_image(&img_vigor_ent_0001, "vigor_ent_0001") ||
      initialize_image(&img_vigor_ent_0002, "vigor_ent_0002") ||
      initialize_image(&img_vigor_ent_0003, "vigor_ent_0003") ||
      initialize_image(&img_vigor_ent_0004, "vigor_ent_0004") ||
      initialize_image(&img_vigor_wav_0001, "vigor_wav_0001") ||
      initialize_image(&img_vigor_wav_0002, "vigor_wav_0002")) {
#if 0
    /* I think this is handled by SDL_Quit. */
    deinitialize_image(img_vigor_empty);
    deinitialize_image(img_vigor_rest);
    deinitialize_image(img_vigor_ent_0001);
    deinitialize_image(img_vigor_ent_0002);
    deinitialize_image(img_vigor_ent_0003);
    deinitialize_image(img_vigor_ent_0004);
    deinitialize_image(img_vigor_wav_0001);
    deinitialize_image(img_vigor_wav_0002);
#endif
    SDL_Quit();
    vigor_vtable.disable_plugin (&vigor_vtable);
    return;
  }

  /* Initialize the speech bubble subsystem. */
  vigor_speech_bubble_init();

  /* Initialize our window */
  screen = SDL_SetVideoMode(img_vigor_rest->w, img_vigor_rest->h,
			    img_vigor_rest->format->BitsPerPixel,
			    SDL_SWSURFACE);
  SDL_WM_SetCaption("Vigor", "Boogie Vigor");

  /* Start worker thread */
  thread_control = GO;
  bv_mutex = SDL_CreateMutex();
  bubble_mutex = SDL_CreateMutex();
  bass_value = 0;
  worker_thread = SDL_CreateThread ((void *) worker_func, NULL);
}

static void
vigor_cleanup(void)
{
  thread_control = STOP;
  vigor_speech_bubble_shutdown(); /* This does nothing presently. :-( */
  SDL_WaitThread(worker_thread, NULL);
  SDL_DestroyMutex(bv_mutex);
  SDL_DestroyMutex(bubble_mutex);
  SDL_Quit();
}

static void
vigor_playback_start(void)
{
  /* FIXME We should move Vigor into frame here. */
}

static void
vigor_playback_stop(void)
{
  /* FIXME We should move Vigor out of frame here. */
  SDL_mutexP(bv_mutex);
  bass_value = 0;
  SDL_mutexV(bv_mutex);
}

#define POPUP_DUR 1
#define POPUP_FREQ 3

static void
vigor_render_freq(gint16 freq_data[2][256])
{
  static time_t last_bubble_time = 0;
  time_t now = time(NULL);
  double dt = now - last_bubble_time;
  static void *last_bubble;

  SDL_mutexP(bv_mutex);
  bass_value = freq_data[0][0];
  SDL_mutexV(bv_mutex);

  /* Sigh.  I'm not supposed to do any intense calculation in here,
   * but since this is the thread in which XMMS is already running
   * GTK, I pretty much have to handle the Vigor quotes here. */

  /* I don't want to use the C library's random function, which
   * isn't reentrant.  Fortunately, I have a decent source of data
   * coming in... */
  if (last_bubble && dt > POPUP_DUR) {
      vigor_speech_bubble_destroy(last_bubble);
      last_bubble = NULL;
  }
  if (!last_bubble && dt > POPUP_FREQ) {
    SDL_SysWMinfo info;
    int res;
    XWindowAttributes attrs;
    int x,y;
    Window dontcare;
    SDL_VERSION(&info.version);
    res = SDL_GetWMInfo(&info);
    if (res == 0) {
      printf("Cant get SDL info");
      exit(0);
    }
    XGetWindowAttributes(info.info.x11.display, info.info.x11.window, &attrs);
    XTranslateCoordinates(info.info.x11.display,
			  info.info.x11.window, attrs.root,
			  attrs.x, attrs.y, &x, &y,
			  &dontcare);
//    printf ("at %d x %d of %d x %d\n", x, y,
//	    attrs.screen->width, attrs.screen->height);
//	    DisplayWidth(info.info.x11.display, attrs.screen),
//	    DisplayHeight(info.info.x11.display, attrs.screen));
    last_bubble = vigor_speech_bubble_create(bass_value,
					     x,y,
					     attrs.screen->width, attrs.screen->height);
    last_bubble_time = now;
  }
}
